"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import re

from PySide6.QtCore import Qt, QEvent
from PySide6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QLabel,
    QListWidget,
    QHBoxLayout,
    QLineEdit,
    QPushButton,
    QListWidgetItem,
    QSizePolicy,
    QSpacerItem,
    QAbstractItemView,
    QMessageBox,
)

from . import config as cfg
from . import utilities as util


class ModifiersList(QDialog):
    """
    class for selection the modifier(s)
    """

    def __init__(self, code: str, modifiers_dict: dict, currentModifier: str):
        super().__init__()
        self.setWindowTitle(cfg.programName)
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

        self.modifiers_dict = dict(modifiers_dict)
        currentModifierList = currentModifier.split("|")

        V1layout = QVBoxLayout()
        label = QLabel()
        label.setText(f"Choose the modifier{'s' * (len(self.modifiers_dict) > 1)} for <b>{code}</b> behavior")
        V1layout.addWidget(label)

        Hlayout = QHBoxLayout()
        self.modifiersSetNumber = 0

        for idx in util.sorted_keys(modifiers_dict):
            if self.modifiers_dict[idx]["type"] not in (
                cfg.SINGLE_SELECTION,
                cfg.MULTI_SELECTION,
                cfg.NUMERIC_MODIFIER,
            ):
                continue

            V2layout = QVBoxLayout()

            self.modifiersSetNumber += 1

            if self.modifiers_dict[idx].get("name", ""):
                lb1 = QLabel(f"Modifier <b>{self.modifiers_dict[idx].get('name', '')}</b>")
                V2layout.addWidget(lb1)

            if self.modifiers_dict[idx].get("description", ""):
                lb2 = QLabel(f"<small>{self.modifiers_dict[idx].get('description', '')[:50]}</small>")
                V2layout.addWidget(lb2)

            if self.modifiers_dict[idx]["type"] in (
                cfg.SINGLE_SELECTION,
                cfg.MULTI_SELECTION,
            ):
                lw = QListWidget()
                self.modifiers_dict[idx]["widget"] = lw
                lw.setObjectName(f"lw_modifiers_({self.modifiers_dict[idx]['type']})")
                lw.installEventFilter(self)

                if self.modifiers_dict[idx]["type"] == cfg.SINGLE_SELECTION:
                    item = QListWidgetItem("None")
                    lw.addItem(item)
                    item.setSelected(True)

                for modifier in self.modifiers_dict[idx]["values"]:
                    item = QListWidgetItem(modifier)
                    if self.modifiers_dict[idx]["type"] == cfg.MULTI_SELECTION:
                        item.setCheckState(Qt.Unchecked)

                        # previously selected
                        try:
                            if currentModifierList != [""] and re.sub(r" \(.\)", "", modifier) in currentModifierList[int(idx)].split(","):
                                item.setCheckState(Qt.Checked)
                        except Exception:  # for old projects due to a fixed bug
                            pass

                    lw.addItem(item)

                    if self.modifiers_dict[idx]["type"] == cfg.SINGLE_SELECTION:
                        try:
                            if currentModifierList != [""] and re.sub(r" \(.\)", "", modifier) == currentModifierList[int(idx)]:
                                item.setSelected(True)
                        except Exception:  # for old projects due to a fixed bug
                            pass
                V2layout.addWidget(lw)

            if self.modifiers_dict[idx]["type"] in [cfg.NUMERIC_MODIFIER]:
                le = QLineEdit()
                self.modifiers_dict[idx]["widget"] = le
                le.setObjectName(f"le_modifiers_({self.modifiers_dict[idx]['type']})")

                if currentModifierList != [""] and currentModifierList[int(idx)] != "None":
                    le.setText(currentModifierList[int(idx)])

                V2layout.addWidget(le)

                # vertical spacer
                spacerItem = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)
                V2layout.addItem(spacerItem)

            Hlayout.addLayout(V2layout)

        V1layout.addLayout(Hlayout)

        H2layout = QHBoxLayout()
        H2layout.addStretch(1)

        pbCancel = QPushButton(cfg.CANCEL)
        pbCancel.clicked.connect(self.reject)
        H2layout.addWidget(pbCancel)

        pbOK = QPushButton(cfg.OK)
        pbOK.setDefault(True)
        pbOK.clicked.connect(self.pbOK_clicked)
        H2layout.addWidget(pbOK)

        V1layout.addLayout(H2layout)
        self.setLayout(V1layout)

        self.installEventFilter(self)
        self.setMaximumSize(1024, 960)

    def eventFilter(self, receiver, event):
        """
        send event (if keypress) to main window
        """
        if event.type() == QEvent.KeyPress:
            ek, ek_text = event.key(), event.text()

            # reject and close dialog if escape pressed
            if ek == Qt.Key_Escape:  # close
                self.reject()
                return False

            # accept and close dialog if enter pressed
            if ek == Qt.Key_Enter or ek == Qt.Key_Return:  # enter or enter from numeric pad
                if not self.pbOK_clicked():
                    return False
                return True

            modifiersSetIndex = 0
            for widget in self.children():
                if "_modifiers" in widget.objectName():
                    modifiersSetIndex = modifiersSetIndex + 1
                if "lw_modifiers" in widget.objectName():
                    if self.modifiersSetNumber == 1:
                        # check if modifiers have code
                        for index in range(widget.count()):
                            if "(" in widget.item(index).text():
                                break
                        else:
                            # modifiers have no associated code: the modifier starting with hit key will be selected
                            if ek not in (Qt.Key_Down, Qt.Key_Up):
                                if ek == Qt.Key_Space and f"({cfg.MULTI_SELECTION})" in widget.objectName():  # checking using SPACE bar
                                    if widget.item(widget.currentRow()).checkState() == Qt.Checked:
                                        widget.item(widget.currentRow()).setCheckState(Qt.Unchecked)
                                    else:
                                        widget.item(widget.currentRow()).setCheckState(Qt.Checked)

                                else:
                                    for index in range(widget.count()):
                                        if widget.item(index).text().upper().startswith(ek_text.upper()):
                                            widget.setCurrentRow(index)
                                            widget.scrollToItem(
                                                widget.item(index),
                                                QAbstractItemView.EnsureVisible,
                                            )
                                            return True
                            else:  # up / down keys
                                try:
                                    if ek == Qt.Key_Down and widget.currentRow() < widget.count() - 1:
                                        widget.setCurrentRow(widget.currentRow() + 1)
                                    if ek == Qt.Key_Up and widget.currentRow() > 0:
                                        widget.setCurrentRow(widget.currentRow() - 1)
                                except Exception:
                                    return

                    for index in range(widget.count()):
                        # check function kesy (F1, F2...)
                        if ek in cfg.function_keys:
                            if f"({cfg.function_keys[ek]})" in widget.item(index).text().upper():
                                if f"({cfg.SINGLE_SELECTION})" in widget.objectName():
                                    widget.item(index).setSelected(True)
                                    # close dialog if one set of modifiers
                                    if self.modifiersSetNumber == 1:
                                        self.accept()
                                        return True
                                    # else move to next set of mofifiers
                                    elif modifiersSetIndex != self.modifiersSetNumber:
                                        widget.parent().focusNextChild()
                                        return True

                                if f"({cfg.MULTI_SELECTION})" in widget.objectName():
                                    if widget.item(index).checkState() == Qt.Checked:
                                        widget.item(index).setCheckState(Qt.Unchecked)
                                    else:
                                        widget.item(index).setCheckState(Qt.Checked)

                        if ek < 1114112 and f"({ek_text})" in widget.item(index).text():
                            if f"({cfg.SINGLE_SELECTION})" in widget.objectName():
                                widget.item(index).setSelected(True)
                                # close dialog if one set of modifiers
                                if self.modifiersSetNumber == 1:
                                    self.accept()
                                    return True
                                # else move to next set of mofifiers
                                elif modifiersSetIndex != self.modifiersSetNumber:
                                    widget.parent().focusNextChild()
                                    return True

                            if f"({cfg.MULTI_SELECTION})" in widget.objectName():
                                if widget.item(index).checkState() == Qt.Checked:
                                    widget.item(index).setCheckState(Qt.Unchecked)
                                else:
                                    widget.item(index).setCheckState(Qt.Checked)

            return True
        else:
            return False

    def get_modifiers(self):
        """
        get modifiers
        returns list of selected modifiers
        """

        for idx in util.sorted_keys(self.modifiers_dict):
            if self.modifiers_dict[idx]["type"] in [
                cfg.SINGLE_SELECTION,
                cfg.MULTI_SELECTION,
                cfg.NUMERIC_MODIFIER,
            ]:
                self.modifiers_dict[idx]["selected"] = []

            if self.modifiers_dict[idx]["type"] == cfg.MULTI_SELECTION:
                for j in range(self.modifiers_dict[idx]["widget"].count()):
                    if self.modifiers_dict[idx]["widget"].item(j).checkState() == Qt.Checked:
                        self.modifiers_dict[idx]["selected"].append(
                            re.sub(
                                r" \(.*\)",
                                "",
                                self.modifiers_dict[idx]["widget"].item(j).text(),
                            )
                        )

                if not self.modifiers_dict[idx]["selected"]:
                    self.modifiers_dict[idx]["selected"].append("None")

            if self.modifiers_dict[idx]["type"] == cfg.SINGLE_SELECTION:
                for item in self.modifiers_dict[idx]["widget"].selectedItems():
                    self.modifiers_dict[idx]["selected"].append(re.sub(r" \(.*\)", "", item.text()))

            if self.modifiers_dict[idx]["type"] == cfg.NUMERIC_MODIFIER:
                self.modifiers_dict[idx]["selected"] = (
                    self.modifiers_dict[idx]["widget"].text() if self.modifiers_dict[idx]["widget"].text() else "None"
                )

        return self.modifiers_dict

    def pbOK_clicked(self):
        """
        OK button is clicked
        """
        for idx in util.sorted_keys(self.modifiers_dict):
            if self.modifiers_dict[idx]["type"] == cfg.NUMERIC_MODIFIER:
                if self.modifiers_dict[idx]["widget"].text():
                    try:
                        _ = float(self.modifiers_dict[idx]["widget"].text())
                    except Exception:
                        QMessageBox.warning(
                            self,
                            cfg.programName,
                            f"<b>{self.modifiers_dict[idx]['widget'].text()}</b> is not a numeric value",
                        )
                        return False

        self.accept()



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging

from PySide6.QtCore import Signal

# from PySide6.QtGui import *
from PySide6.QtWidgets import (
    QApplication,
    QWidget,
    QRadioButton,
    QLabel,
    QHBoxLayout,
    QVBoxLayout,
    QLineEdit,
    QPlainTextEdit,
    QCheckBox,
    QPushButton,
    QFileDialog,
    QMessageBox,
)
from . import dialog
from . import config as cfg


class wgMeasurement(QWidget):
    """ """

    closeSignal, clearSignal = Signal(), Signal()
    flagSaved = True
    draw_mem = []

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Geometric measurements")

        vbox = QVBoxLayout(self)

        self.rbPoint = QRadioButton("Point (left click)")
        vbox.addWidget(self.rbPoint)

        self.rbDistance = QRadioButton("Distance (start: left click, end: right click)")
        vbox.addWidget(self.rbDistance)

        self.rbArea = QRadioButton("Area (left click for area vertices, right click to close area)")
        vbox.addWidget(self.rbArea)

        self.rbAngle = QRadioButton("Angle (vertex: left click, segments: right click)")
        vbox.addWidget(self.rbAngle)

        self.cbPersistentMeasurements = QCheckBox("Measurements are persistent")
        self.cbPersistentMeasurements.setChecked(True)
        vbox.addWidget(self.cbPersistentMeasurements)

        vbox.addWidget(QLabel("<b>Scale</b>"))

        hbox1 = QHBoxLayout()

        self.lbRef = QLabel("Reference")
        hbox1.addWidget(self.lbRef)

        self.lbPx = QLabel("Pixels")
        hbox1.addWidget(self.lbPx)

        vbox.addLayout(hbox1)

        hbox2 = QHBoxLayout()

        self.leRef = QLineEdit()
        self.leRef.setText("1")
        hbox2.addWidget(self.leRef)

        self.lePx = QLineEdit()
        self.lePx.setText("1")
        hbox2.addWidget(self.lePx)

        vbox.addLayout(hbox2)

        self.pte = QPlainTextEdit()
        vbox.addWidget(self.pte)

        self.status_lb = QLabel()
        vbox.addWidget(self.status_lb)

        hbox3 = QHBoxLayout()

        self.pbClear = QPushButton("Clear measurements", clicked=self.pbClear_clicked)
        hbox3.addWidget(self.pbClear)

        self.pbSave = QPushButton("Save results", clicked=self.pbSave_clicked)
        hbox3.addWidget(self.pbSave)

        self.pbClose = QPushButton(cfg.CLOSE, clicked=self.pbClose_clicked)
        hbox3.addWidget(self.pbClose)

        vbox.addLayout(hbox3)

    def pbClear_clicked(self):
        """
        clear measurements draw and results
        """
        self.draw_mem = {}
        self.pte.clear()
        self.clearSignal.emit()

    def pbClose_clicked(self):
        if not self.flagSaved:
            response = dialog.MessageDialog(
                cfg.programName,
                "The current results are not saved. Do you want to save results before closing?",
                [cfg.YES, cfg.NO, cfg.CANCEL],
            )
            if response == cfg.YES:
                self.pbSave_clicked()
            if response == cfg.CANCEL:
                return
        self.closeSignal.emit()

    def pbSave_clicked(self):
        """
        save results
        """
        if self.pte.toPlainText():
            fileName, _ = QFileDialog().getSaveFileName(self, "Save measurement results", "", "Text files (*.txt);;All files (*)")
            if fileName:
                with open(fileName, "w") as f:
                    f.write(self.pte.toPlainText())
                self.flagSaved = True
        else:
            QMessageBox.information(self, cfg.programName, "There are no results to save")


if __name__ == "__main__":
    import sys

    app = QApplication(sys.argv)
    w = wgMeasurement(logging.getLogger().getEffectiveLevel())
    w.show()

    sys.exit(app.exec_())



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.


Parse an OTX/ODX or OTB (compressed OTX) file and convert the ethogram, modifiers, subjects
and independent variables to BORIS format

"""

import datetime as dt
from decimal import Decimal as dec
import re
import zipfile
import pathlib as pl
from xml.dom import minidom
import logging
from typing import Tuple

try:
    from . import config as cfg
except Exception:
    import config as cfg


def otx_to_boris(file_path: str) -> Tuple[dict, list]:
    """
    convert otx/otb/odx file in a BORIS project

    For ODX files ask to import observations

    Args:
        file_path (str): path to otx/otb/odx file

    Returns:
        dict: BORIS project
        list: list of errors during importation
    """

    if pl.Path(file_path).suffix == ".otb":
        with zipfile.ZipFile(file_path) as file_zip:
            files_list = file_zip.namelist()
            if files_list:
                try:
                    file_zip.extract(files_list[0])
                except Exception:
                    return {"fatal": True}, ["Error when extracting file from OTB"]
            else:
                return {"fatal": True}, ["Error when extracting file"]

            try:
                xmldoc = minidom.parse(files_list[0])
            except Exception:
                return {"fatal": True}, ["XML parsing error"]

    elif pl.Path(file_path).suffix in (".odx", ".otx"):
        try:
            xmldoc = minidom.parse(file_path)
        except Exception:
            return {"fatal": True}, ["XML parsing error"]

    else:
        return {"fatal": True}, ["The file must be in OTB, OTX or ODX format"]

    flag_long_key: bool = False
    error_list: list = []

    # metadata
    for item in xmldoc.getElementsByTagName("MET_METADATA"):
        metadata = minidom.parseString(item.toxml())
        try:
            project_name = re.sub("<[^>]*>", "", metadata.getElementsByTagName("MET_PROJECT_NAME")[0].toxml())
        except Exception:
            project_name = ""
        try:
            project_description = re.sub("<[^>]*>", "", metadata.getElementsByTagName("MET_PROJECT_DESCRIPTION")[0].toxml())
        except Exception:
            project_description = ""

        try:
            project_creation_date = re.sub("<[^>]*>", "", metadata.getElementsByTagName("MET_CREATION_DATETIME")[0].toxml())
        except Exception:
            project_creation_date = ""

    # modifiers
    modifiers: dict = {}
    # modifiers_set = {}
    itemlist = xmldoc.getElementsByTagName("CDS_MODIFIER")
    for item in itemlist:
        modif = minidom.parseString(item.toxml())

        modif_code = re.sub("<[^>]*>", "", modif.getElementsByTagName("CDS_ELE_NAME")[0].toxml())

        modif_id = re.sub("<[^>]*>", "", modif.getElementsByTagName("CDS_ELE_ID")[0].toxml())

        try:
            modif_parent_id = re.sub("<[^>]*>", "", modif.getElementsByTagName("CDS_ELE_PARENT_ID")[0].toxml())
        except Exception:
            modif_parent_id = ""

        try:
            description = re.sub("<[^>]*>", "", modif.getElementsByTagName("CDS_ELE_DESCRIPTION")[0].toxml())
        except Exception:
            description = ""
        try:
            key = re.sub("<[^>]*>", "", modif.getElementsByTagName("CDS_ELE_START_KEYCODE")[0].toxml())
        except Exception:
            key = ""

        if modif_parent_id:
            modifiers[modif_parent_id]["values"].append(modif_code)
        else:
            if len(key) > 1:
                key = ""
                flag_long_key = True
            modifiers[modif_id] = {"set_name": modif_code, "key": key, "description": description, "values": []}

    logging.debug(modifiers)

    # connect modifiers to behaviors
    connections: dict = {}
    itemlist = xmldoc.getElementsByTagName("CDS_CONNECTION")
    for item in itemlist:
        if item.attributes["CDS_ELEMENT_ID"].value not in connections:
            connections[item.attributes["CDS_ELEMENT_ID"].value] = []
        connections[item.attributes["CDS_ELEMENT_ID"].value].append(item.attributes["CDS_MODIFIER_ID"].value)

    logging.debug(connections)

    # behaviors
    behaviors: dict = {}
    behaviors_list: list = []
    behav_category: list = []
    mutually_exclusive_list: list = []
    itemlist = xmldoc.getElementsByTagName("CDS_BEHAVIOR")
    for item in itemlist:
        behav = minidom.parseString(item.toxml())

        behav_code = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_NAME")[0].toxml())

        behav_id = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_ID")[0].toxml())

        try:
            description = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_DESCRIPTION")[0].toxml())
        except Exception:
            description = ""
        try:
            key = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_START_KEYCODE")[0].toxml())
        except Exception:
            key = ""

        try:
            stop_key = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_STOP_KEYCODE")[0].toxml())
        except Exception:
            stop_key = ""

        try:
            parent_name = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_PARENT_NAME")[0].toxml())
        except Exception:
            parent_name = ""

        try:
            mutually_exclusive = re.sub("<[^>]*>", "", behav.getElementsByTagName("CDS_ELE_MUT_EXCLUSIVE")[0].toxml())
        except Exception:
            mutually_exclusive = ""

        if mutually_exclusive == "Y" and parent_name:
            mutually_exclusive_list.append(behav_code)

        if behav_id in connections:
            modifier_sets = [modifiers[modifier_set]["set_name"] for modifier_set in connections[behav_id]]
        else:
            modifier_sets = []

        if parent_name:  # behavior
            if (not key or len(key) > 1) and stop_key:
                key = stop_key

            if len(key) > 1:
                key = ""
                flag_long_key = True

            behaviors[str(len(behaviors))] = {
                "id": int(behav_id),
                "code": behav_code,
                "key": key,
                "description": description,
                "modifiers": modifier_sets,
                "category": parent_name,
            }
            behaviors_list.append(behav_code)

        else:  #  behavioral category
            behav_category.append(behav_code)

    behaviors_boris: dict = {}
    for k in behaviors:
        behaviors_boris[k] = {
            "code": behaviors[k]["code"],
            "type": cfg.POINT_EVENT,
            "key": behaviors[k]["key"],
            "description": behaviors[k]["description"],
            "category": behaviors[k]["category"],
            "excluded": "",
            "coding map": "",
        }

        if behaviors[k]["code"] in mutually_exclusive_list:
            behaviors_boris[k]["excluded"] = ",".join([x for x in behaviors_list if x != behaviors[k]["code"]])

        behaviors_boris[k]["modifiers"] = {}
        if behaviors[k]["modifiers"]:
            for modif_key in modifiers:
                if modifiers[modif_key]["set_name"] in behaviors[k]["modifiers"]:
                    new_index = str(len(behaviors_boris[k]["modifiers"]))
                    behaviors_boris[k]["modifiers"][new_index] = {
                        "name": modifiers[modif_key]["set_name"],
                        "type": cfg.SINGLE_SELECTION,
                        "values": modifiers[modif_key]["values"],
                        "description": modifiers[modif_key]["description"],
                    }

    logging.debug(behaviors_boris)

    # subjects
    subjects = {}
    itemlist = xmldoc.getElementsByTagName("CDS_SUBJECT")
    for item in itemlist:
        subject = minidom.parseString(item.toxml())
        subject_name = re.sub("<[^>]*>", "", subject.getElementsByTagName("CDS_ELE_NAME")[0].toxml())
        try:
            key = re.sub("<[^>]*>", "", subject.getElementsByTagName("CDS_ELE_START_KEYCODE")[0].toxml())
        except Exception:
            key = ""
        try:
            parent_name = re.sub("<[^>]*>", "", subject.getElementsByTagName("CDS_ELE_PARENT_NAME")[0].toxml())
        except Exception:
            parent_name = ""

        if parent_name:
            if len(key) > 1:
                key = ""
                flag_long_key = True
            subjects[str(len(subjects))] = {"key": key, "name": subject_name, "description": ""}

    # independent variables
    variables = {}
    itemlist = xmldoc.getElementsByTagName("VL_VARIABLE")
    for item in itemlist:
        variable = minidom.parseString(item.toxml())

        variable_label = re.sub("<[^>]*>", "", variable.getElementsByTagName("VL_LABEL")[0].toxml())

        variable_id = re.sub("<[^>]*>", "", variable.getElementsByTagName("VL_ID")[0].toxml())

        variable_type = re.sub("<[^>]*>", "", variable.getElementsByTagName("VL_TYPE")[0].toxml())
        if variable_type.upper() == "TEXT":
            variable_type = cfg.TEXT
        if variable_type.upper() == "DOUBLE":
            variable_type = cfg.NUMERIC
        if variable_type.upper() == "FILEREFERENCE":
            variable_type = cfg.TEXT
        if variable_type.upper() == "DURATION":
            variable_type = cfg.TEXT
        if variable_type.upper() == "TIMESTAMP":
            variable_type = cfg.TIMESTAMP
        if variable_type.upper() == "BOOLEAN":
            variable_type = cfg.TEXT

        try:
            variable_description = re.sub("<[^>]*>", "", modif.getElementsByTagName("VL_DESCRIPTION")[0].toxml())
        except Exception:
            variable_description = ""

        try:
            values = variable.getElementsByTagName("VL_VALUE")
            values_list = []
            for value in values:
                values_list.append(re.sub("<[^>]*>", "", value.toxml()))
            values_str = ",".join(values_list)

        except Exception:
            values_str = ""

        variables[variable_id] = {
            "label": variable_label,
            "type": variable_type.lower(),
            "description": variable_description,
        }
        if values_str:
            variables[variable_id]["predefined_values"] = values_str
            variables[variable_id]["type"] = "value from set"

    variables_boris = {}
    for k in variables:
        variables_boris[k] = {
            "label": variables[k]["label"],
            "description": variables[k]["description"],
            "type": variables[k]["type"],
            "default value": "",
            "possible values": variables[k]["predefined_values"] if "predefined_values" in variables[k] else "",
        }

    # create empty project from template
    project = dict(cfg.EMPTY_PROJECT)
    project[cfg.OBSERVATIONS] = {}

    observations = xmldoc.getElementsByTagName("OBS_OBSERVATION")

    for OBS_OBSERVATION in observations:
        # OBS_OBSERVATION = minidom.parseString(OBS_OBSERVATION.toxml())

        obs_id = OBS_OBSERVATION.getAttribute("NAME")

        project[cfg.OBSERVATIONS][obs_id] = dict(
            {
                "file": {},
                "type": "LIVE",
                "description": "",
                "time offset": 0,
                cfg.EVENTS: [],
                "observation time interval": [0, 0],
                "independent_variables": {},
                "visualize_spectrogram": False,
                "visualize_waveform": False,
                "close_behaviors_between_videos": False,
                "scan_sampling_time": 0,
                "start_from_current_time": False,
                "start_from_current_epoch_time": False,
            }
        )

        OBS_EVENT_LOGS = OBS_OBSERVATION.getElementsByTagName("OBS_EVENT_LOGS")[0]

        for OBS_EVENT_LOG in OBS_EVENT_LOGS.getElementsByTagName("OBS_EVENT_LOG"):
            CREATION_DATETIME = OBS_EVENT_LOG.getAttribute("CREATION_DATETIME")

            CREATION_DATETIME = CREATION_DATETIME.replace(" ", "T")  # .split(".")[0]

            logging.debug(f"{CREATION_DATETIME=}")  # ex: 2022-05-18 10:04:09.474512"""

            project[cfg.OBSERVATIONS][obs_id]["date"] = CREATION_DATETIME

            for event in OBS_EVENT_LOG.getElementsByTagName("OBS_EVENT"):
                OBS_EVENT_TIMESTAMP = event.getElementsByTagName("OBS_EVENT_TIMESTAMP")[0].childNodes[0].data

                full_timestamp = dt.datetime.strptime(OBS_EVENT_TIMESTAMP, "%Y-%m-%d %H:%M:%S.%f").timestamp()
                logging.debug(f"{full_timestamp=}")

                # day_timestamp = dt.datetime.strptime(OBS_EVENT_TIMESTAMP.split(" ")[0], "%Y-%m-%d").timestamp()
                # timestamp = dec(str(round(full_timestamp - day_timestamp, 3)))
                timestamp = dec(full_timestamp).quantize(dec(".001"))

                try:
                    OBS_EVENT_SUBJECT = event.getElementsByTagName("OBS_EVENT_SUBJECT")[0].getAttribute("NAME")
                except Exception:
                    OBS_EVENT_SUBJECT = ""

                OBS_EVENT_BEHAVIOR = event.getElementsByTagName("OBS_EVENT_BEHAVIOR")[0].getAttribute("NAME")
                logging.debug(f"{OBS_EVENT_BEHAVIOR=}")
                if not OBS_EVENT_BEHAVIOR:
                    logging.warning(f"Behavior missing in observation {obs_id} at {timestamp}")
                    error_list.append(f"Behavior missing in observation {obs_id} at {timestamp}")
                    continue

                # modifier
                try:
                    OBS_EVENT_BEHAVIOR_MODIFIER = (
                        event.getElementsByTagName("OBS_EVENT_BEHAVIOR")[0]
                        .getElementsByTagName("OBS_EVENT_BEHAVIOR_MODIFIER")[0]
                        .childNodes[0]
                        .data
                    )
                except Exception:
                    OBS_EVENT_BEHAVIOR_MODIFIER: str = ""

                # comment
                try:
                    OBS_EVENT_COMMENT: str = event.getElementsByTagName("OBS_EVENT_COMMENT")[0].childNodes[0].data
                except Exception:
                    OBS_EVENT_COMMENT: str = ""

                logging.debug(f"{timestamp=}")
                logging.debug(f"{OBS_EVENT_SUBJECT=}")
                logging.debug(f"{OBS_EVENT_BEHAVIOR=}")
                logging.debug(f"{OBS_EVENT_BEHAVIOR_MODIFIER=}")
                logging.debug(f"{OBS_EVENT_COMMENT=}")

                project[cfg.OBSERVATIONS][obs_id][cfg.EVENTS].append(
                    [
                        timestamp,
                        OBS_EVENT_SUBJECT,
                        OBS_EVENT_BEHAVIOR,
                        OBS_EVENT_BEHAVIOR_MODIFIER,
                        OBS_EVENT_COMMENT,
                    ]
                )

    project[cfg.PROJECT_NAME] = project_name
    project[cfg.PROJECT_DATE] = project_creation_date.replace(" ", "T")
    project[cfg.ETHOGRAM] = behaviors_boris
    project[cfg.PROJECT_DESCRIPTION] = project_description
    project[cfg.BEHAVIORAL_CATEGORIES] = behav_category
    project[cfg.SUBJECTS] = subjects
    project[cfg.INDEPENDENT_VARIABLES] = variables_boris

    if flag_long_key:
        error_list.append("The keys longer than one char were deleted.")
        logging.debug("The keys longer than one char were deleted.")

    return project, error_list


if __name__ == "__main__":
    import sys
    import pprint

    logging.basicConfig(
        format="%(asctime)s,%(msecs)d  %(module)s l.%(lineno)d %(levelname)s %(message)s",
        datefmt="%H:%M:%S",
        level=logging.DEBUG,
    )
    project, errors = otx_to_boris(sys.argv[1])

    pprint.pprint(project)
    pprint.pprint(errors)



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import datetime as dt
from decimal import Decimal as dec
import logging
import math
import pathlib as pl
import platform
import sys
import traceback
from typing import Union

from PySide6.QtCore import Qt, Signal, qVersion, QRect, QTime, QDateTime, QSize
from PySide6.QtWidgets import (
    QApplication,
    QAbstractItemView,
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QPlainTextEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSpinBox,
    QDoubleSpinBox,
    QTableView,
    QTableWidget,
    QVBoxLayout,
    QWidget,
    QDateTimeEdit,
    QTimeEdit,
    QAbstractSpinBox,
    QRadioButton,
    QStackedWidget,
    QFrame,
)
from PySide6.QtGui import QFont, QTextCursor

from . import config as cfg
from . import version
from . import utilities as util


def MessageDialog(title: str, text: str, buttons: tuple) -> str:
    """
    show a generic message dialog and returns the text of the clicked button

    Args:
        title (str): Title of the dialog box
        text (str): text of the dialog box
        buttons (tuple): text for buttons

    Return
        str: text of the clicked button
    """
    message = QMessageBox()
    message.setWindowTitle(title)
    message.setText(text)
    message.setIcon(QMessageBox.Question)
    for button in buttons:
        message.addButton(button, QMessageBox.YesRole)

    message.setWindowFlags(message.windowFlags() | Qt.WindowStaysOnTopHint)
    message.exec()
    return message.clickedButton().text()


def global_error_message(exception_type, exception_value, traceback_object):
    """
    Global error management
    save error using loggin.critical and stdout
    """

    error_text = "\n\nSystem info\n===========\n\n"
    error_text += util.get_systeminfo()
    error_text += f"Error succeded at {dt.datetime.now():%Y-%m-%d %H:%M}\n\n"
    error_text += "".join(traceback.format_exception(exception_type, exception_value, traceback_object))

    # write to stdout
    logging.critical(error_text)

    # write to $HOME/boris_error.log
    try:
        with open(pl.Path.home() / "boris_error.log", "w") as f_error:
            f_error.write(error_text)
    except Exception:
        logging.critical(f"Impossible to write to {pl.Path.home() / 'boris_error.log'}")

    # copy to clipboard
    cb = QApplication.clipboard()
    cb.clear()
    cb.setText(error_text)

    text: str = (
        f"An error has occured!\n\n"
        "to improve the software please report this problem at:\n"
        "https://github.com/olivierfriard/BORIS/issues\n\n"
        "Please no screenshot, the error message was copied to the clipboard.\n\n"
        "Thank you for your collaboration!\n\n"
        f"{error_text}"
    )

    errorbox = Results_dialog()

    errorbox.setWindowTitle("BORIS - An error occured")
    errorbox.pbOK.setText("Abort")
    errorbox.pbCancel.setVisible(True)
    errorbox.pbCancel.setText("Ignore and try to continue")

    font = QFont()
    font.setFamily("monospace")
    errorbox.ptText.setFont(font)
    errorbox.ptText.clear()
    errorbox.ptText.appendPlainText(text)

    errorbox.ptText.moveCursor(QTextCursor.Start)

    ret = errorbox.exec_()

    if ret == 1:  # Abort
        sys.exit(1)


class Info_widget(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle("BORIS")
        layout = QVBoxLayout()
        self.label = QLabel()
        layout.addWidget(self.label)
        self.lwi = QListWidget()
        layout.addWidget(self.lwi)
        self.setLayout(layout)


class get_time_widget(QWidget):
    """
    widget for selecting a time in various formats: seconds, HH:MM:SS:ZZZ or YYYY-mm-DD HH:MM:SS:ZZZ
    """

    def __init__(self, time_value=dec(0), parent=None):
        super().__init__(parent)

        self.setWindowTitle("BORIS")

        self.widget = QWidget()
        self.widget.setObjectName("widget")
        self.widget.setGeometry(QRect(130, 220, 302, 63))
        self.verticalLayout_3 = QVBoxLayout(self.widget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.label = QLabel(self.widget)
        self.label.setObjectName("label")

        self.verticalLayout_2.addWidget(self.label)

        self.pb_sign = QPushButton(self.widget)
        self.pb_sign.setObjectName("pb_sign")
        self.pb_sign.setMaximumSize(QSize(40, 16777215))

        self.verticalLayout_2.addWidget(self.pb_sign)

        self.horizontalLayout_3.addLayout(self.verticalLayout_2)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.rb_seconds = QRadioButton(self.widget)
        self.rb_seconds.setObjectName("rb_seconds")

        self.horizontalLayout.addWidget(self.rb_seconds)

        self.rb_time = QRadioButton(self.widget)
        self.rb_time.setObjectName("rb_time")

        self.horizontalLayout.addWidget(self.rb_time)

        self.rb_datetime = QRadioButton(self.widget)
        self.rb_datetime.setObjectName("rb_datetime")

        self.horizontalLayout.addWidget(self.rb_datetime)

        self.horizontalLayout.addStretch()

        self.verticalLayout.addLayout(self.horizontalLayout)

        self.stackedWidget = QStackedWidget(self.widget)
        self.stackedWidget.setObjectName("stackedWidget")
        self.stackedWidget.setFrameShape(QFrame.NoFrame)
        self.seconds = QWidget()
        self.seconds.setObjectName("seconds")
        self.widget1 = QWidget(self.seconds)
        self.widget1.setObjectName("widget1")
        # self.widget1.setGeometry(QRect(10, 0, 163, 27))
        self.horizontalLayout_4 = QHBoxLayout(self.widget1)
        self.horizontalLayout_4.setObjectName("horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.le_seconds = QLineEdit(self.widget1)
        self.le_seconds.setObjectName("le_seconds")

        self.horizontalLayout_4.addWidget(self.le_seconds)

        self.lb_seconds = QLabel(self.widget1)
        self.lb_seconds.setObjectName("lb_seconds")

        self.horizontalLayout_4.addWidget(self.lb_seconds)

        self.horizontalLayout_4.addStretch()

        self.stackedWidget.addWidget(self.seconds)
        self.hhmmss = QWidget()
        self.hhmmss.setObjectName("hhmmss")
        self.widget2 = QWidget(self.hhmmss)
        self.widget2.setMinimumWidth(500)
        self.widget2.setObjectName("widget2")
        self.widget2.setGeometry(QRect(0, 0, 213, 28))
        self.horizontalLayout_2 = QHBoxLayout(self.widget2)
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.lb_hour = QLabel(self.widget2)
        self.lb_hour.setObjectName("lb_hour")

        self.horizontalLayout_2.addWidget(self.lb_hour)

        self.sb_hour = QSpinBox(self.widget2)
        self.sb_hour.setObjectName("sb_hour")
        self.sb_hour.setButtonSymbols(QAbstractSpinBox.NoButtons)

        self.horizontalLayout_2.addWidget(self.sb_hour)

        self.lb_hhmmss = QLabel(self.widget2)
        self.lb_hhmmss.setObjectName("lb_hhmmss")

        self.horizontalLayout_2.addWidget(self.lb_hhmmss)

        self.te_time = QTimeEdit(self.widget2)
        self.te_time.setObjectName("te_time")
        self.te_time.adjustSize()
        # self.te_time.setMinimumWidth(200)
        # self.widget2.adjustSize()

        self.horizontalLayout_2.addWidget(self.te_time)

        self.horizontalLayout_2.addStretch()

        self.stackedWidget.addWidget(self.hhmmss)
        self.page_2 = QWidget()
        self.page_2.setObjectName("page_2")
        self.dte = QDateTimeEdit(self.page_2)

        self.dte.setObjectName("dte")
        # self.dte.setGeometry(QRect(10, 0, 164, 26))
        self.stackedWidget.addWidget(self.page_2)

        self.verticalLayout.addWidget(self.stackedWidget)

        self.horizontalLayout_3.addLayout(self.verticalLayout)

        self.verticalLayout_3.addLayout(self.horizontalLayout_3)

        self.line = QFrame(self.widget)
        self.line.setObjectName("line")
        self.line.setFrameShape(QFrame.HLine)
        self.line.setFrameShadow(QFrame.Sunken)

        self.verticalLayout_3.addWidget(self.line)

        self.setLayout(self.verticalLayout_3)

        self.stackedWidget.setCurrentIndex(0)

        self.label.setText("")
        self.pb_sign.setText("+")
        self.rb_seconds.setText("Seconds")
        self.rb_time.setText("hh:mm:ss")
        self.rb_datetime.setText("Date time")
        self.le_seconds.setText("")
        self.lb_seconds.setText("seconds")
        self.lb_hour.setText("hour")
        self.lb_hhmmss.setText("mm:ss.ms")
        self.te_time.setDisplayFormat("mm:ss.zzz")
        self.dte.setDisplayFormat("yyyy-MM-dd hh:mm:ss:zzz")
        self.dte.adjustSize()
        font = QFont()
        font.setPointSize(14)
        self.pb_sign.setFont(font)

        self.rb_seconds.toggled.connect(self.format_changed)
        self.rb_time.toggled.connect(self.format_changed)
        self.rb_datetime.toggled.connect(self.format_changed)
        self.sb_hour.setMaximum(cfg.HOUR_CUTOFF)
        self.pb_sign.clicked.connect(self.pb_sign_clicked)

        if time_value:
            self.set_time(time_value)

        self.format_changed()

        self.adjustSize()

    def format_changed(self):
        if self.rb_seconds.isChecked():
            self.stackedWidget.setCurrentIndex(0)
        if self.rb_time.isChecked():
            self.stackedWidget.setCurrentIndex(1)
        if self.rb_datetime.isChecked():
            self.stackedWidget.setCurrentIndex(2)

        self.le_seconds.setEnabled(self.rb_seconds.isChecked())
        self.le_seconds.adjustSize()
        self.lb_seconds.setEnabled(self.rb_seconds.isChecked())
        self.sb_hour.setEnabled(self.rb_time.isChecked())
        self.te_time.setEnabled(self.rb_time.isChecked())
        self.lb_hour.setEnabled(self.rb_time.isChecked())
        self.lb_hhmmss.setEnabled(self.rb_time.isChecked())
        self.dte.setEnabled(self.rb_datetime.isChecked())

    def pb_sign_clicked(self):
        if self.pb_sign.text() == "+":
            self.pb_sign.setText("-")
        else:
            self.pb_sign.setText("+")

    def get_time(self) -> Union[dec, None]:
        """
        Get time from the selected format in the time widget

        Returns:
            dec: time in seconds (None if no format selected)
        """

        time_sec = dec("NaN")

        if self.rb_seconds.isChecked():
            try:
                time_sec = float(self.le_seconds.text())
            except Exception:
                QMessageBox.warning(
                    None,
                    cfg.programName,
                    f"The value of seconds ({self.le_seconds.text()}) is not a decimal number",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )
                return dec("NaN")

        if self.rb_time.isChecked():
            time_sec = self.sb_hour.value() * 3600
            time_sec += self.te_time.time().msecsSinceStartOfDay() / 1000

        if self.rb_datetime.isChecked():
            time_sec = self.dte.dateTime().toMSecsSinceEpoch() / 1000

        if self.pb_sign.text() == "-":
            time_sec = -time_sec

        return dec(time_sec).quantize(dec("0.001"))  # if time_sec is not None else None

    def set_time(self, new_time: dec) -> None:
        """
        set time on time widget
        """

        if math.isnan(new_time):
            return

        self.pb_sign.setText("-" if new_time < 0 else "+")

        # seconds
        self.le_seconds.setText(f"{new_time:0.3f}")

        if new_time <= cfg.DATE_CUTOFF:  # hh:mm:ss.zzz
            h = int(abs(new_time) // 3600)
            m = int((abs(new_time) - h * 3600) // 60)
            s = int((abs(new_time) - h * 3600 - m * 60))
            ms = round((abs(new_time) - h * 3600 - m * 60 - s) * 1000)

            self.sb_hour.setValue(h)
            self.te_time.setTime(QTime(0, m, s, ms))
        else:
            self.sb_hour.setValue(0)
            self.te_time.setTime(QTime(0, 0, 0, 0))

            self.dte.setDateTime(QDateTime().fromMSecsSinceEpoch(int(new_time * 1000)))
            self.rb_datetime.setChecked(True)


class Ask_time(QDialog):
    """
    Qdialog class for asking time to user
    User can select a time format between seconds, HHMMSS.zzz or YYY-mm-DD HH:MM:SS.zzz
    """

    def __init__(self, time_value=0):
        super().__init__()
        self.setWindowTitle("")

        hbox = QVBoxLayout(self)
        self.label = QLabel()
        self.label.setText("")
        hbox.addWidget(self.label)

        self.time_widget = get_time_widget(time_value)

        hbox.addWidget(self.time_widget)

        self.pbOK = QPushButton(cfg.OK, clicked=self.pb_ok_clicked)
        self.pbOK.setDefault(True)

        self.pbCancel = QPushButton(cfg.CANCEL)
        self.pbCancel.clicked.connect(self.reject)

        self.hbox2 = QHBoxLayout(self)
        self.hbox2.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        self.hbox2.addWidget(self.pbCancel)
        self.hbox2.addWidget(self.pbOK)
        hbox.addLayout(self.hbox2)
        self.setLayout(hbox)

        # if time_value:
        #    self.time_widget.set_time(time_value)

    def pb_ok_clicked(self):
        if (
            not self.time_widget.rb_seconds.isChecked()
            and not self.time_widget.rb_time.isChecked()
            and not self.time_widget.rb_datetime.isChecked()
        ):
            QMessageBox.warning(
                None,
                cfg.programName,
                "Select an option",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return
        # test time value
        if self.time_widget.get_time().is_nan():
            return

        self.accept()


class Video_overlay_dialog(QDialog):
    """
    dialog to ask image file and position for video overlay
    """

    def __init__(self):
        super().__init__()

        vlayout = QVBoxLayout()
        self.cb_player = QComboBox()
        vlayout.addWidget(self.cb_player)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Image file"))
        self.le_file_path = QLineEdit()
        hbox.addWidget(self.le_file_path)
        vlayout.addLayout(hbox)

        hbox = QHBoxLayout()
        self.pb_browse = QPushButton("Browse", self, clicked=self.browse)
        hbox.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        hbox.addWidget(self.pb_browse)
        vlayout.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Position: x,y"))
        self.le_overlay_position = QLineEdit()
        hbox.addWidget(self.le_overlay_position)
        vlayout.addLayout(hbox)

        hbox = QHBoxLayout()
        hbox.addWidget(QLabel("Transparency %"))
        self.sb_overlay_transparency = QSpinBox()
        self.sb_overlay_transparency.setRange(0, 100)
        self.sb_overlay_transparency.setSingleStep(1)
        self.sb_overlay_transparency.setValue(90)
        hbox.addWidget(self.sb_overlay_transparency)
        vlayout.addLayout(hbox)

        # self.sb_overlay_transparency.setEnabled(False)

        hbox = QHBoxLayout()
        self.pb_cancel = QPushButton("Cancel", self, clicked=self.reject)
        hbox.addWidget(self.pb_cancel)
        self.pb_OK = QPushButton("OK", clicked=self.ok)

        self.pb_OK.setDefault(True)
        hbox.addWidget(self.pb_OK)

        vlayout.addLayout(hbox)

        self.setLayout(vlayout)

    def browse(self):
        file_name, _ = QFileDialog.getOpenFileName(self, "Choose an image file", "", "PNG files (*.png);;All files (*)")
        if file_name:
            self.le_file_path.setText(file_name)

    def ok(self):
        if not self.le_file_path.text():
            QMessageBox.warning(
                None,
                cfg.programName,
                "Select a file containing a PNG image",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return

        if self.le_overlay_position.text() and "," not in self.le_overlay_position.text():
            QMessageBox.warning(
                None,
                cfg.programName,
                "The overlay position must be in x,y format",
                QMessageBox.Ok | QMessageBox.Default,
                QMessageBox.NoButton,
            )
            return
        if self.le_overlay_position.text():
            try:
                [int(x.strip()) for x in self.le_overlay_position.text().split(",")]
            except Exception:
                QMessageBox.warning(
                    None,
                    cfg.programName,
                    "The overlay position must be in x,y format",
                    QMessageBox.Ok | QMessageBox.Default,
                    QMessageBox.NoButton,
                )
                return
        self.accept()


class Input_dialog(QDialog):
    """
    dialog for user input. Elements can be:
        checkbox (cb): Tuple(str, str, bool)
        lineedit (le): Tuple(str, str)
        spinbox (sb)
        doubleSpinbox (dsb)
        items list (il)

    """

    def __init__(self, label_caption: str, elements_list: list, title: str = ""):
        super().__init__()

        self.setWindowTitle(title)

        hbox = QVBoxLayout()
        self.label = QLabel()
        self.label.setText(label_caption)
        hbox.addWidget(self.label)

        self.elements: dict = {}
        for element in elements_list:
            if element[0] == "cb":  # checkbox
                self.elements[element[1]] = QCheckBox(element[1])
                self.elements[element[1]].setChecked(element[2])
                hbox.addWidget(self.elements[element[1]])

            if element[0] == "le":  # line edit
                lb = QLabel(element[1])
                hbox.addWidget(lb)
                self.elements[element[1]] = QLineEdit()
                hbox.addWidget(self.elements[element[1]])

            if element[0] == "sb":  # spinbox
                # 1 - Label
                # 2 - minimum value
                # 3 - maximum value
                # 4 - step
                # 5 - initial value

                lb = QLabel(element[1])
                hbox.addWidget(lb)
                self.elements[element[1]] = QSpinBox()
                self.elements[element[1]].setRange(element[2], element[3])
                self.elements[element[1]].setSingleStep(element[4])
                self.elements[element[1]].setValue(element[5])
                hbox.addWidget(self.elements[element[1]])

            if element[0] == "dsb":  # doubleSpinbox
                # 1 - Label
                # 2 - minimum value
                # 3 - maximum value
                # 4 - step
                # 5 - initial value
                # 6 - number of decimals

                lb = QLabel(element[1])
                hbox.addWidget(lb)
                self.elements[element[1]] = QDoubleSpinBox()
                self.elements[element[1]].setRange(element[2], element[3])
                self.elements[element[1]].setSingleStep(element[4])
                self.elements[element[1]].setValue(element[5])
                self.elements[element[1]].setDecimals(element[6])
                hbox.addWidget(self.elements[element[1]])

            if element[0] == "il":  # items list
                # 1 - Label
                # 2 - Values (tuple of tuple: 0 - value; 1 - "", "selected")
                lb = QLabel(element[1])
                hbox.addWidget(lb)
                self.elements[element[1]] = QComboBox()
                self.elements[element[1]].addItems([x[0] for x in element[2]])  # take first element of tuple
                try:
                    self.elements[element[1]].setCurrentIndex([idx for idx, x in enumerate(element[2]) if x[1] == "selected"][0])
                except Exception:
                    self.elements[element[1]].setCurrentIndex(0)
                hbox.addWidget(self.elements[element[1]])

        hbox2 = QHBoxLayout()

        hbox2.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))

        self.pbCancel = QPushButton(cfg.CANCEL)
        self.pbCancel.clicked.connect(self.reject)
        hbox2.addWidget(self.pbCancel)

        self.pbOK = QPushButton(cfg.OK, clicked=self.accept)
        self.pbOK.setDefault(True)
        hbox2.addWidget(self.pbOK)

        hbox.addLayout(hbox2)

        self.setLayout(hbox)


class Duplicate_items(QDialog):
    """
    let user show between behaviors/subjects that are coded by same key
    """

    def __init__(self, text, codes_list):
        super(Duplicate_items, self).__init__()

        self.setWindowTitle(cfg.programName)
        self.setWindowFlags(Qt.WindowStaysOnTopHint)

        Vlayout = QVBoxLayout()
        widget = QWidget(self)
        widget.setLayout(Vlayout)

        label = QLabel()
        label.setText(text)
        Vlayout.addWidget(label)

        self.lw = QListWidget(widget)
        self.lw.setObjectName("lw_modifiers")
        # TODO: to be enabled
        # lw.installEventFilter(self)

        for code in codes_list:
            item = QListWidgetItem(code)
            self.lw.addItem(item)

        Vlayout.addWidget(self.lw)

        hlayout = QHBoxLayout()
        hlayout.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        pbCancel = QPushButton("Cancel", clicked=self.reject)
        hlayout.addWidget(pbCancel)

        pbOK = QPushButton("OK", clicked=self.accept)
        pbOK.setDefault(True)
        hlayout.addWidget(pbOK)

        Vlayout.addLayout(hlayout)

        self.setLayout(Vlayout)

        # self.installEventFilter(self)

        self.setMaximumSize(1024, 960)

    def getCode(self):
        """
        get selected behavior code
        """
        if self.lw.selectedItems():
            return self.lw.selectedItems()[0].text()
        else:
            return None


class ChooseObservationsToImport(QDialog):
    """
    dialog for selectiong items
    """

    def __init__(self, text, observations_list):
        super(ChooseObservationsToImport, self).__init__()

        self.setWindowTitle(cfg.programName)

        Vlayout = QVBoxLayout()
        widget = QWidget(self)
        widget.setLayout(Vlayout)

        label = QLabel()
        label.setText(text)
        Vlayout.addWidget(label)

        self.lw = QListWidget(widget)
        self.lw.setObjectName("lw_observations")
        self.lw.setSelectionMode(QAbstractItemView.ExtendedSelection)
        # TODO: to be enabled
        # lw.installEventFilter(self)

        for code in observations_list:
            item = QListWidgetItem(code)
            self.lw.addItem(item)

        Vlayout.addWidget(self.lw)

        hlayout = QHBoxLayout()
        hlayout.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))
        pbCancel = QPushButton("Cancel", clicked=self.reject)
        hlayout.addWidget(pbCancel)
        pbOK = QPushButton("OK", clicked=self.accept)
        pbOK.setDefault(True)
        hlayout.addWidget(pbOK)

        Vlayout.addLayout(hlayout)

        self.setLayout(Vlayout)

        # self.installEventFilter(self)

        self.setMaximumSize(1024, 960)

    def get_selected_observations(self):
        """
        get selected_observations
        """
        return [item.text() for item in self.lw.selectedItems()]


class FindInEvents(QWidget):
    """
    "find in events" dialog box
    """

    clickSignal = Signal(str)
    currentIdx: int = -1

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Find in events")

        hbox = QVBoxLayout()

        self.cbSubject = QCheckBox("Subject")
        self.cbSubject.setChecked(True)
        hbox.addWidget(self.cbSubject)

        self.cbBehavior = QCheckBox("Behavior")
        self.cbBehavior.setChecked(True)
        hbox.addWidget(self.cbBehavior)

        self.cbModifier = QCheckBox("Modifiers")
        self.cbModifier.setChecked(True)
        hbox.addWidget(self.cbModifier)

        self.cbComment = QCheckBox("Comment")
        self.cbComment.setChecked(True)
        hbox.addWidget(self.cbComment)

        self.lbFind = QLabel("Find")
        hbox.addWidget(self.lbFind)

        self.findText = QLineEdit()
        hbox.addWidget(self.findText)

        self.cbFindInSelectedEvents = QCheckBox("Find in selected events")
        self.cbFindInSelectedEvents.setChecked(False)
        hbox.addWidget(self.cbFindInSelectedEvents)

        self.cb_case_sensitive = QCheckBox("Case sensitive")
        self.cb_case_sensitive.setChecked(False)
        hbox.addWidget(self.cb_case_sensitive)

        self.lb_message = QLabel()
        hbox.addWidget(self.lb_message)

        hbox2 = QHBoxLayout()
        hbox2.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))
        self.pbOK = QPushButton("Find")
        self.pbOK.clicked.connect(lambda: self.click("FIND"))
        self.pbCancel = QPushButton(cfg.CLOSE)
        self.pbCancel.clicked.connect(lambda: self.click("CLOSE"))
        hbox2.addWidget(self.pbCancel)
        hbox2.addWidget(self.pbOK)
        hbox.addLayout(hbox2)

        self.setLayout(hbox)

    def click(self, msg):
        self.clickSignal.emit(msg)


class FindReplaceEvents(QWidget):
    """
    "find replace events" dialog box
    """

    clickSignal = Signal(str)

    def __init__(self):
        super().__init__()

        self.setWindowTitle("Find/Replace events")

        hbox = QVBoxLayout()

        self.combo_fields = QComboBox()
        self.combo_fields.addItems(("Choose a field", "Subject", "Behavior", "Modifiers", "Comment"))
        hbox.addWidget(self.combo_fields)

        self.lbFind = QLabel("Find")
        hbox.addWidget(self.lbFind)

        self.findText = QLineEdit()
        hbox.addWidget(self.findText)

        self.lbReplace = QLabel("Replace")
        hbox.addWidget(self.lbReplace)

        self.replaceText = QLineEdit()
        hbox.addWidget(self.replaceText)

        self.cbFindInSelectedEvents = QCheckBox("Find/Replace only in selected events")
        self.cbFindInSelectedEvents.setChecked(False)
        hbox.addWidget(self.cbFindInSelectedEvents)

        self.cb_case_sensitive = QCheckBox("Case sensitive")
        self.cb_case_sensitive.setChecked(False)
        hbox.addWidget(self.cb_case_sensitive)

        hbox2 = QHBoxLayout()

        hbox2.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))

        self.pbCancel = QPushButton(cfg.CANCEL, clicked=lambda: self.click("CANCEL"))
        hbox2.addWidget(self.pbCancel)

        self.pbOK = QPushButton("Find and replace", clicked=lambda: self.click("FIND_REPLACE"))
        hbox2.addWidget(self.pbOK)

        self.pbFindReplaceAll = QPushButton("Find and replace all", clicked=lambda: self.click("FIND_REPLACE_ALL"))
        hbox2.addWidget(self.pbFindReplaceAll)

        hbox.addLayout(hbox2)

        self.lb_results = QLabel(" ")
        hbox.addWidget(self.lb_results)

        self.setLayout(hbox)

    def click(self, msg):
        self.clickSignal.emit(msg)


class Results_dialog(QDialog):
    """
    widget for visualizing text output in HTML
    """

    def __init__(self):
        super().__init__()

        self.dataset = False

        self.setWindowTitle("")

        hbox = QVBoxLayout()

        self.lb = QLabel("")
        hbox.addWidget(self.lb)

        self.ptText = QPlainTextEdit()
        self.ptText.setReadOnly(True)
        hbox.addWidget(self.ptText)

        hbox2 = QHBoxLayout()
        hbox2.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.pbSave = QPushButton("Save results", clicked=self.save_results)
        hbox2.addWidget(self.pbSave)

        self.pbCancel = QPushButton(cfg.CANCEL, clicked=self.reject)
        hbox2.addWidget(self.pbCancel)
        self.pbCancel.setVisible(False)

        self.pbOK = QPushButton(cfg.OK, clicked=self.accept)
        hbox2.addWidget(self.pbOK)

        hbox.addLayout(hbox2)
        self.setLayout(hbox)

        self.resize(800, 640)

    def save_results(self):
        """
        save content of self.ptText
        """

        if not self.dataset:
            file_name, _ = QFileDialog().getSaveFileName(self, "Save results", "", "Text files (*.txt *.tsv);;All files (*)")

            if not file_name:
                return
            try:
                with open(file_name, "w") as f:
                    f.write(self.ptText.toPlainText())
            except Exception:
                QMessageBox.critical(self, cfg.programName, f"The file {file_name} can not be saved")

        else:
            self.done(cfg.SAVE_DATASET)


class Results_dialog_exit_code(QDialog):
    """
    widget for visualizing text output
    """

    def __init__(self):
        super().__init__()

        self.dataset = False

        self.setWindowTitle("")

        hbox = QVBoxLayout()

        self.lb = QLabel("")
        hbox.addWidget(self.lb)

        self.ptText = QPlainTextEdit()
        self.ptText.setReadOnly(True)
        hbox.addWidget(self.ptText)

        hbox2 = QHBoxLayout()
        hbox2.addItem(QSpacerItem(241, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.pbSave = QPushButton("Save results", clicked=self.save_results)
        hbox2.addWidget(self.pbSave)

        self.pb1 = QPushButton("1", clicked=lambda: self.done_(1))
        hbox2.addWidget(self.pb1)

        self.pb2 = QPushButton("2", clicked=lambda: self.done_(2))
        hbox2.addWidget(self.pb2)

        self.pb3 = QPushButton("3", clicked=lambda: self.done_(3))
        hbox2.addWidget(self.pb3)

        hbox.addLayout(hbox2)
        self.setLayout(hbox)

        self.resize(800, 640)

    def done_(self, status):
        self.done(status)

    def save_results(self):
        """
        save content of self.ptText
        """

        if not self.dataset:
            file_name, _ = QFileDialog().getSaveFileName(self, "Save results", "", "Text files (*.txt *.tsv);;All files (*)")

            if not file_name:
                return
            try:
                with open(file_name, "w") as f:
                    f.write(self.ptText.toPlainText())
            except Exception:
                QMessageBox.critical(self, cfg.programName, f"The file {file_name} can not be saved")

        else:
            self.done(cfg.SAVE_DATASET)


class View_data(QDialog):
    """
    widget for visualizing rows of data file
    """

    def __init__(self):
        super().__init__()

        self.setWindowTitle("")

        vbox = QVBoxLayout()

        self.lb = QLabel("")
        vbox.addWidget(self.lb)

        self.tw = QTableWidget()
        self.tw.verticalHeader().hide()
        vbox.addWidget(self.tw)

        vbox.addWidget(QLabel("Descriptive statistics"))

        self.stats = QPlainTextEdit()
        font = QFont()
        font.setFamily("Monospace")
        self.stats.setFont(font)

        vbox.addWidget(self.stats)

        self.label = QLabel("Enter the column indices to plot (time, value) separated by comma (,)")
        vbox.addWidget(self.label)

        self.le = QLineEdit()
        vbox.addWidget(self.le)

        hbox2 = QHBoxLayout()

        hbox2.addItem(QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum))

        self.pbCancel = QPushButton(cfg.CANCEL, clicked=self.reject)
        hbox2.addWidget(self.pbCancel)

        self.pbOK = QPushButton(cfg.OK, clicked=self.accept)
        hbox2.addWidget(self.pbOK)

        vbox.addLayout(hbox2)

        self.setLayout(vbox)

        self.resize(800, 640)


class View_explore_project_results(QWidget):
    """
    widget for visualizing results of explore project
    """

    double_click_signal = Signal(str, int)

    def __init__(self):
        super().__init__()

        self.setWindowTitle("")

        vbox = QVBoxLayout()

        self.lb = QLabel("")
        vbox.addWidget(self.lb)

        self.tw = QTableWidget()
        self.tw.setSelectionBehavior(QTableView.SelectRows)
        self.tw.cellDoubleClicked[int, int].connect(self.tw_cellDoubleClicked)
        vbox.addWidget(self.tw)

        hbox2 = QHBoxLayout()
        hbox2.addItem(QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum))
        hbox2.addWidget(QPushButton(cfg.CLOSE, clicked=self.close))

        vbox.addLayout(hbox2)

        self.setLayout(vbox)

    def tw_cellDoubleClicked(self, r, c):
        self.double_click_signal.emit(self.tw.item(r, 0).text(), int(self.tw.item(r, 1).text()))



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import logging
import pathlib as pl

from PySide6.QtWidgets import QFileDialog, QMessageBox
from PySide6.QtGui import QFont, QTextOption, QTextCursor

from . import config as cfg
from . import (
    dialog,
    project_functions,
    select_observations,
    select_subj_behav,
    time_budget_functions,
    observation_operations,
)


def synthetic_time_budget(self) -> None:
    """
    Synthetic time budget
    """

    logging.debug("synthetic time budget function")

    _, selected_observations = select_observations.select_observations2(
        self, mode=cfg.MULTIPLE, windows_title="Select observations for synthetic time budget"
    )

    if not selected_observations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    max_media_duration_all_obs, _ = observation_operations.media_duration(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_interval, end_interval = observation_operations.time_intervals_range(self.pj[cfg.OBSERVATIONS], selected_observations)

    synth_tb_param = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=start_coding,
        end_coding=end_coding,
        # start_interval=start_interval,
        # end_interval=end_interval,
        start_interval=None,
        end_interval=None,
        maxTime=max_media_duration_all_obs,
        show_exclude_non_coded_behaviors=False,
        by_category=False,
        n_observations=len(selected_observations),
        show_exclude_non_coded_modifiers=True,
    )

    if synth_tb_param == {}:
        return

    if not synth_tb_param[cfg.SELECTED_SUBJECTS] or not synth_tb_param[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # ask for excluding behaviors durations from total time
    if not start_coding.is_nan():
        cancel_pressed, synth_tb_param[cfg.EXCLUDED_BEHAVIORS] = self.filter_behaviors(
            title="Select behaviors to exclude from the total time",
            text="The duration of the selected behaviors will be subtracted from the total time",
            table="",
            behavior_type=cfg.STATE_EVENT_TYPES,
        )
        if cancel_pressed:
            return
    else:
        synth_tb_param[cfg.EXCLUDED_BEHAVIORS] = []

    ok, msg, data_report = time_budget_functions.synthetic_time_budget(self.pj, selected_observations, synth_tb_param)

    results = dialog.Results_dialog()
    results.setWindowTitle("Synthetic time budget")
    if not ok:
        results.ptText.clear()
        results.ptText.setReadOnly(True)
        results.ptText.appendHtml(msg.replace("\n", "<br>"))
        results.exec_()
        return

    results.dataset = True
    font = QFont("Courier", 12)
    results.ptText.setFont(font)
    results.ptText.setWordWrapMode(QTextOption.NoWrap)
    results.ptText.setReadOnly(True)
    results.ptText.appendPlainText(data_report.export("cli", tablefmt="grid"))  # other available format: github
    results.ptText.moveCursor(QTextCursor.Start)
    results.resize(960, 640)

    if results.exec_() == cfg.SAVE_DATASET:
        file_formats = [
            cfg.TSV,
            cfg.CSV,
            cfg.ODS,
            cfg.XLSX,
            cfg.XLS,
            cfg.HTML,
            cfg.TEXT_FILE,  # tablib format: cli
        ]

        file_name, filter_ = QFileDialog().getSaveFileName(self, "Synthetic time budget", "", ";;".join(file_formats))
        if not file_name:
            return

        output_format = cfg.FILE_NAME_SUFFIX[filter_]

        if pl.Path(file_name).suffix != "." + output_format:
            if filter_ != cfg.TEXT_FILE:
                file_name = str(pl.Path(file_name)) + "." + output_format
            else:
                file_name = str(pl.Path(file_name))
            if pl.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", [cfg.CANCEL, cfg.OVERWRITE])
                    == cfg.CANCEL
                ):
                    return

        with open(file_name, "wb") as f:
            if filter_ in (cfg.TSV, cfg.CSV, cfg.HTML, cfg.TEXT_FILE):
                f.write(str.encode(data_report.export(output_format)))
            if filter_ in (cfg.ODS, cfg.XLSX, cfg.XLS):
                f.write(data_report.export(output_format))


def synthetic_binned_time_budget(self) -> None:
    """
    Synthetic time budget with time bin
    """

    _, selected_observations = select_observations.select_observations2(
        self, mode=cfg.MULTIPLE, windows_title="Select observations for synthetic binned time budget"
    )

    if not selected_observations:
        return

    # check if coded behaviors are defined in ethogram
    if project_functions.check_coded_behaviors_in_obs_list(self.pj, selected_observations):
        return

    # check if state events are paired
    not_ok, selected_observations = project_functions.check_state_events(self.pj, selected_observations)
    if not_ok or not selected_observations:
        return

    max_media_duration_all_obs, total_media_duration_all_obs = observation_operations.media_duration(
        self.pj[cfg.OBSERVATIONS], selected_observations
    )

    start_coding, end_coding, _ = observation_operations.coding_time(self.pj[cfg.OBSERVATIONS], selected_observations)

    start_interval, end_interval = observation_operations.time_intervals_range(self.pj[cfg.OBSERVATIONS], selected_observations)

    # exit with message if events do not have timestamp
    if start_coding.is_nan():
        QMessageBox.critical(
            None,
            cfg.programName,
            ("This function is not available for observations with events that do not have timestamp"),
            QMessageBox.Ok | QMessageBox.Default,
            QMessageBox.NoButton,
        )
        return

    synth_tb_param = select_subj_behav.choose_obs_subj_behav_category(
        self,
        selected_observations,
        start_coding=start_coding,
        end_coding=end_coding,
        # start_interval=start_interval,
        # end_interval=end_interval,
        start_interval=None,
        end_interval=None,
        maxTime=max_media_duration_all_obs,
        show_exclude_non_coded_behaviors=False,
        by_category=False,
        n_observations=len(selected_observations),
        show_time_bin_size=True,
        show_exclude_non_coded_modifiers=True,
    )

    if synth_tb_param == {}:
        return

    if not synth_tb_param[cfg.SELECTED_SUBJECTS] or not synth_tb_param[cfg.SELECTED_BEHAVIORS]:
        QMessageBox.warning(None, cfg.programName, "Select subject(s) and behavior(s) to analyze")
        return

    # ask for excluding behaviors durations from total time
    cancel_pressed, synth_tb_param[cfg.EXCLUDED_BEHAVIORS] = self.filter_behaviors(
        title="Select behaviors to exclude",
        text=("The duration of the selected behaviors will be subtracted from the total time"),
        table="",
        behavior_type=cfg.STATE_EVENT_TYPES,
    )
    if cancel_pressed:
        return

    ok, data_report = time_budget_functions.synthetic_time_budget_bin(self.pj, selected_observations, synth_tb_param)

    if not ok:
        results = dialog.Results_dialog()
        results.setWindowTitle("Synthetic time budget with time bin")
        results.ptText.appendHtml("Error during the creation of the synthetic time budget with time bin")
        results.exec_()
        return

    results = dialog.Results_dialog()
    results.dataset = True
    results.setWindowTitle("Synthetic time budget by time bin")
    font = QFont("Courier", 12)
    results.ptText.setFont(font)
    results.ptText.setWordWrapMode(QTextOption.NoWrap)
    results.ptText.setReadOnly(True)
    results.ptText.appendPlainText(data_report.export("cli", tablefmt="grid"))  # other available format: github
    results.ptText.moveCursor(QTextCursor.Start)
    results.resize(960, 640)

    if results.exec_() == cfg.SAVE_DATASET:
        file_formats = [
            cfg.TSV,
            cfg.CSV,
            cfg.ODS,
            cfg.XLSX,
            cfg.XLS,
            cfg.HTML,
            cfg.TEXT_FILE,
        ]

        file_name, filter_ = QFileDialog().getSaveFileName(
            self, "Save the Synthetic time budget with time bin", "", ";;".join(file_formats)
        )
        if not file_name:
            return

        output_format = cfg.FILE_NAME_SUFFIX[filter_]

        if pl.Path(file_name).suffix != "." + output_format:
            if filter_ != cfg.TEXT_FILE:
                file_name = str(pl.Path(file_name)) + "." + output_format
            else:
                file_name = str(pl.Path(file_name))
            if pl.Path(file_name).is_file():
                if (
                    dialog.MessageDialog(cfg.programName, f"The file {file_name} already exists.", (cfg.CANCEL, cfg.OVERWRITE))
                    == cfg.CANCEL
                ):
                    return

        with open(file_name, "wb") as f:
            if filter_ in (cfg.TSV, cfg.CSV, cfg.HTML, cfg.TEXT_FILE):
                f.write(str.encode(data_report.export(output_format)))
            if filter_ in (cfg.ODS, cfg.XLSX, cfg.XLS):
                f.write(data_report.export(output_format))



"""
BORIS plugin

Time budget
"""

import pandas as pd
import numpy as np

__version__ = "0.3.0"
__version_date__ = "2025-03-17"
__plugin_name__ = "Time budget"
__author__ = "Olivier Friard - University of Torino - Italy"


def run(df: pd.DataFrame):
    """
    Calculate the following values:

    - Total number of occurences of behavior
    - Total duration of behavior (in seconds)  (pandas.DataFrame.sum() ignore NaN values when computing the sum. Use min_count=1)
    - Duration mean of behavior (in seconds)
    - Standard deviation of behavior duration (in seconds)
    - Inter-event intervals mean (in seconds)
    - Inter-event intervals standard deviation (in seconds)
    - % of total subject observation duration
    """

    print("running time budget plugin")

    print(df)

    group_by = ["Subject", "Behavior"]

    dfs = [
        df.groupby(group_by)["Behavior"].count().reset_index(name="number of occurences"),
        df.groupby(group_by)["Duration (s)"].sum(min_count=1).reset_index(name="total duration"),
        df.groupby(group_by)["Duration (s)"].mean().astype(float).round(3).reset_index(name="duration mean"),
        df.groupby(group_by)["Duration (s)"].std().astype(float).round(3).reset_index(name="duration std dev"),
    ]

    # inter events
    df2 = df.sort_values(by=["Observation id", "Subject", "Behavior", "Start (s)"])
    df2["diff"] = df2.groupby(["Observation id", "Subject", "Behavior"])["Start (s)"].shift(periods=-1) - df2["Stop (s)"]

    dfs.append(df2.groupby(group_by)["diff"].mean().astype(float).round(3).reset_index(name="inter-event intervals mean"))

    dfs.append(df2.groupby(group_by)["diff"].std().astype(float).round(3).reset_index(name="inter-event intervals std dev"))

    # % of total subject observation time

    interval = (df.groupby(["Subject"])["Stop (s)"].max() - df.groupby(["Subject"])["Start (s)"].min()).replace(0, np.nan)

    dfs.append(
        (100 * df.groupby(group_by)["Duration (s)"].sum(min_count=1) / interval)
        .astype(float)
        .round(3)
        .reset_index(name="% of total subject observation duration")
    )

    merged_df = dfs[0]
    for df in dfs[1:]:
        merged_df = pd.merge(merged_df, df, on=group_by)

    return merged_df



"""
BORIS
Behavioral Observation Research Interactive Software
Copyright 2012-2025 Olivier Friard

This file is part of BORIS.

  BORIS is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 3 of the License, or
  any later version.

  BORIS is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not see <http://www.gnu.org/licenses/>.

"""

import datetime as dt
import pathlib as pl

import matplotlib

matplotlib.use("Qt5Agg")

import matplotlib.dates

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.dates import (
    DateFormatter,
)

from . import config as cfg
from . import db_functions, project_functions, observation_operations
from . import utilities as util

# matplotlib.pyplot.switch_backend("Qt5Agg")


def default_value(ethogram, behavior, parameter):
    """
    return value for duration in case of point event
    """
    default_value_ = 0
    if project_functions.event_type(behavior, ethogram) in cfg.POINT_EVENT_TYPES and parameter in ["duration"]:
        default_value_ = "NA"
    return default_value_


def init_behav_modif(
    ethogram: dict,
    selected_subjects: list,
    distinct_behav_modif,
    include_modifiers,
    parameters,
) -> dict:
    """
    initialize dictionary with subject, behaviors and modifiers
    """

    behaviors = {}
    for subj in selected_subjects:
        behaviors[subj] = {}
        for behav_modif in distinct_behav_modif:
            behav, modif = behav_modif
            behav_modif_str = "|".join(behav_modif) if modif else behav

            if behav_modif_str not in behaviors[subj]:
                behaviors[subj][behav_modif_str] = {}

            for parameter in parameters:
                behaviors[subj][behav_modif_str][parameter[0]] = default_value(ethogram, behav_modif_str, parameter[0])

    return behaviors


def init_behav(ethogram: dict, selected_subjects: list, distinct_behaviors, parameters) -> dict:
    """
    initialize dictionary with subject, behaviors and modifiers
    """

    behaviors: dict = {}
    for subj in selected_subjects:
        behaviors[subj] = {}
        for behavior in distinct_behaviors:
            if behavior not in behaviors[subj]:
                behaviors[subj][behavior] = {}
            for parameter in parameters:
                behaviors[subj][behavior][parameter] = default_value(ethogram, behavior, parameter)
    return behaviors


def create_behaviors_bar_plot(
    pj: dict,
    selected_observations: list,
    param: dict,
    plot_directory: str,
    output_format: str,
    plot_colors: list = cfg.BEHAVIORS_PLOT_COLORS,
) -> dict:
    """
    time budget bar plot

    Args:
        pj (dict): project
        param (dict): parameters
        plot_directory (str): path of directory
        output_format (str): image format

    Returns:
        dict:
    """

    selected_subjects = param[cfg.SELECTED_SUBJECTS]
    selected_behaviors = param[cfg.SELECTED_BEHAVIORS]
    start_time = param[cfg.START_TIME]
    end_time = param[cfg.END_TIME]

    parameters = ["duration", "number of occurences"]

    ok, msg, db_connector = db_functions.load_aggregated_events_in_db(pj, selected_subjects, selected_observations, selected_behaviors)

    if not ok:
        return {"error": True, "message": msg}

    # extract all behaviors from ethogram for colors in plot
    all_behaviors = util.all_behaviors(pj[cfg.ETHOGRAM])

    for obs_id in selected_observations:
        cursor = db_connector.cursor()
        # distinct behaviors
        cursor.execute(
            "SELECT distinct behavior FROM aggregated_events WHERE observation = ?",
            (obs_id,),
        )
        distinct_behav = [rows["behavior"] for rows in cursor.fetchall()]

        # add selected behaviors that are not observed
        """
        if not param[EXCLUDE_BEHAVIORS]:
            for behavior in selected_behaviors:
                if [x for x in distinct_behav if x == behavior] == []:
                    distinct_behav.append(behavior)
        """

        # distinct subjects
        cursor.execute(
            "SELECT distinct subject FROM aggregated_events WHERE observation = ?",
            (obs_id,),
        )
        distinct_subjects = [rows["subject"] for rows in cursor.fetchall()]

        behaviors = init_behav(pj[cfg.ETHOGRAM], distinct_subjects, distinct_behav, parameters)

        # plot creation
        if len(distinct_subjects) > 1:
            fig, axs = plt.subplots(nrows=1, ncols=len(distinct_subjects), sharey=True)
            fig2, axs2 = plt.subplots(nrows=1, ncols=len(distinct_subjects), sharey=True)

        else:
            fig, ax = plt.subplots(nrows=1, ncols=len(distinct_subjects), sharey=True)
            axs = np.ndarray(shape=(1), dtype=type(ax))
            axs[0] = ax

            fig2, ax2 = plt.subplots(nrows=1, ncols=len(distinct_subjects), sharey=True)
            axs2 = np.ndarray(shape=(1), dtype=type(ax2))
            axs2[0] = ax2

        fig.suptitle("Durations of behaviors")
        fig2.suptitle("Number of occurences of behaviors")

        # if modifiers not to be included set modifiers to ""
        cursor.execute("UPDATE aggregated_events SET modifiers = ''")

        # time
        obs_length = observation_operations.observation_total_length(pj[cfg.OBSERVATIONS][obs_id])
        if obs_length == -1:
            obs_length = 0

        if param["time"] == cfg.TIME_FULL_OBS:
            min_time = float(0)
            max_time = float(obs_length)

        if param["time"] == cfg.TIME_EVENTS:
            try:
                min_time = float(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][0][0])
            except Exception:
                min_time = float(0)
            try:
                max_time = float(pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0])
            except Exception:
                max_time = float(obs_length)

        if param["time"] in (cfg.TIME_ARBITRARY_INTERVAL, cfg.TIME_OBS_INTERVAL):
            min_time = float(start_time)
            max_time = float(end_time)

        cursor.execute(
            "UPDATE aggregated_events SET start = ? WHERE observation = ? AND start < ? AND stop BETWEEN ? AND ?",
            (
                min_time,
                obs_id,
                min_time,
                min_time,
                max_time,
            ),
        )
        cursor.execute(
            "UPDATE aggregated_events SET stop = ? WHERE observation = ? AND stop > ? AND start BETWEEN ? AND ?",
            (
                max_time,
                obs_id,
                max_time,
                min_time,
                max_time,
            ),
        )
        cursor.execute(
            "UPDATE aggregated_events SET start = ?, stop = ? WHERE observation = ? AND start < ? AND stop > ?",
            (
                min_time,
                max_time,
                obs_id,
                min_time,
                max_time,
            ),
        )

        for ax_idx, subject in enumerate(sorted(distinct_subjects)):
            for behavior in distinct_behav:
                # number of occurences
                cursor.execute(
                    ("SELECT COUNT(*) AS count FROM aggregated_events WHERE observation = ? AND subject = ? AND behavior = ?"),
                    (
                        obs_id,
                        subject,
                        behavior,
                    ),
                )
                for row in cursor.fetchall():
                    behaviors[subject][behavior]["number of occurences"] = 0 if row["count"] is None else row["count"]

                # total duration
                if project_functions.event_type(behavior, pj[cfg.ETHOGRAM]) in cfg.STATE_EVENT_TYPES:
                    cursor.execute(
                        (
                            "SELECT SUM(stop - start) AS duration FROM aggregated_events "
                            "WHERE observation = ? AND subject = ? AND behavior = ?"
                        ),
                        (
                            obs_id,
                            subject,
                            behavior,
                        ),
                    )
                    for row in cursor.fetchall():
                        behaviors[subject][behavior]["duration"] = 0 if row["duration"] is None else row["duration"]

            (
                durations,
                n_occurences,
                colors,
                x_labels,
                colors_duration,
                x_labels_duration,
            ) = ([], [], [], [], [], [])

            for behavior in sorted(distinct_behav):
                if param[cfg.EXCLUDE_BEHAVIORS] and behaviors[subject][behavior]["number of occurences"] == 0:
                    continue

                n_occurences.append(behaviors[subject][behavior]["number of occurences"])
                x_labels.append(behavior)

                # color
                behav_idx = [k for k in pj[cfg.ETHOGRAM] if pj[cfg.ETHOGRAM][k]["code"] == behavior][0]
                col = None
                if cfg.COLOR in pj[cfg.ETHOGRAM][behav_idx]:
                    col = util.behavior_user_color(pj[cfg.ETHOGRAM], behavior)
                if col is not None:
                    colors.append(col)
                else:
                    try:
                        colors.append(util.behavior_color(plot_colors, all_behaviors.index(behavior)))
                    except Exception:
                        colors.append("darkgray")

                if project_functions.event_type(behavior, pj[cfg.ETHOGRAM]) in cfg.STATE_EVENT_TYPES:
                    durations.append(behaviors[subject][behavior]["duration"])
                    x_labels_duration.append(behavior)

                    col = None
                    if cfg.COLOR in pj[cfg.ETHOGRAM][behav_idx]:
                        col = util.behavior_user_color(pj[cfg.ETHOGRAM], behavior)
                    if col is not None:
                        colors_duration.append(col)
                    else:
                        try:
                            colors_duration.append(util.behavior_color(plot_colors, all_behaviors.index(behavior)))
                        except Exception:
                            colors_duration.append("darkgray")

            # width = 0.35       # the width of the bars: can also be len(x) sequence

            axs2[ax_idx].bar(
                np.arange(len(n_occurences)),
                n_occurences,
                # width,
                color=colors,
            )

            axs[ax_idx].bar(
                np.arange(len(durations)),
                durations,
                # width,
                color=colors_duration,
            )

            if ax_idx == 0:
                axs[ax_idx].set_ylabel("Duration (s)")
            axs[ax_idx].set_xlabel("Behaviors")
            axs[ax_idx].set_title(f"{subject}")

            axs[ax_idx].set_xticks(np.arange(len(durations)))
            axs[ax_idx].set_xticklabels(x_labels_duration, rotation="vertical", fontsize=8)

            if ax_idx == 0:
                axs2[ax_idx].set_ylabel("Number of occurences")
            axs2[ax_idx].set_xlabel("Behaviors")
            axs2[ax_idx].set_title(f"{subject}")

            axs2[ax_idx].set_xticks(np.arange(len(n_occurences)))
            axs2[ax_idx].set_xticklabels(x_labels, rotation="vertical", fontsize=8)

        fig.align_labels()
        fig.tight_layout(rect=[0, 0.03, 1, 0.95])

        fig2.align_labels()
        fig2.tight_layout(rect=[0, 0.03, 1, 0.95])

        if plot_directory:
            # output_file_name = f"{pathlib.Path(plot_directory) / utilities.safeFileName(obs_id)}.{output_format}"
            fig.savefig(f"{pl.Path(plot_directory) / util.safeFileName(obs_id)}.duration.{output_format}")
            fig2.savefig(f"{pl.Path(plot_directory) / util.safeFileName(obs_id)}.number_of_occurences.{output_format}")
            plt.close()
        else:
            fig.show()
            fig2.show()

    return {}


def create_events_plot(
    self,
    selected_observations,
    parameters,
    plot_colors=cfg.BEHAVIORS_PLOT_COLORS,
    plot_directory="",
    file_format="png",
):
    """
    create a time diagram plot (like a gantt chart)
    with matplotlib barh function (https://matplotlib.org/3.1.0/api/_as_gen/matplotlib.pyplot.barh.html)
    """

    selected_subjects = parameters[cfg.SELECTED_SUBJECTS]
    selected_behaviors = parameters[cfg.SELECTED_BEHAVIORS]
    include_modifiers = parameters[cfg.INCLUDE_MODIFIERS]
    interval = parameters[cfg.TIME_INTERVAL]
    start_time = parameters[cfg.START_TIME]
    end_time = parameters[cfg.END_TIME]

    ok, msg, db_connector = db_functions.load_aggregated_events_in_db(self.pj, selected_subjects, selected_observations, selected_behaviors)

    if not ok:
        return False, msg, None

    cursor = db_connector.cursor()

    # if modifiers not to be included set modifiers to ""
    if not include_modifiers:
        cursor.execute("UPDATE aggregated_events SET modifiers = ''")

    cursor.execute("SELECT DISTINCT behavior, modifiers FROM aggregated_events")
    distinct_behav_modif = [[rows["behavior"], rows["modifiers"]] for rows in cursor.fetchall()]

    # add selected behaviors that are not observed
    for behav in selected_behaviors:
        if [x for x in distinct_behav_modif if x[0] == behav] == []:
            distinct_behav_modif.append([behav, "-"])

    distinct_behav_modif = sorted(distinct_behav_modif)
    max_len = len(distinct_behav_modif)

    all_behaviors = [self.pj[cfg.ETHOGRAM][x][cfg.BEHAVIOR_CODE] for x in util.sorted_keys(self.pj[cfg.ETHOGRAM])]

    par1 = 1
    bar_height = 0.5
    epoch_date = dt.datetime(2017, 1, 1)

    for obs_id in selected_observations:
        if len(selected_subjects) > 1:
            fig, axs = plt.subplots(figsize=(20, 8), nrows=len(selected_subjects), ncols=1, sharex=True)
        else:
            fig, ax = plt.subplots(figsize=(20, 8), nrows=len(selected_subjects), ncols=1, sharex=True)
            axs = np.ndarray(shape=(1), dtype=type(ax))
            axs[0] = ax

        ok, msg, db_connector = db_functions.load_aggregated_events_in_db(self.pj, selected_subjects, [obs_id], selected_behaviors)

        cursor = db_connector.cursor()
        # if modifiers not to be included set modifiers to ""
        if not include_modifiers:
            cursor.execute("UPDATE aggregated_events SET modifiers = ''")
        cursor = db_connector.cursor()

        cursor.execute("SELECT DISTINCT behavior, modifiers FROM aggregated_events")
        distinct_behav_modif = [[rows["behavior"], rows["modifiers"]] for rows in cursor.fetchall()]

        # add selected behaviors that are not observed
        if not parameters["exclude behaviors"]:
            for behav in selected_behaviors:
                if [x for x in distinct_behav_modif if x[0] == behav] == []:
                    distinct_behav_modif.append([behav, "-"])

        distinct_behav_modif = sorted(distinct_behav_modif)
        max_len = len(distinct_behav_modif)

        # time
        obs_length = observation_operations.observation_total_length(self.pj[cfg.OBSERVATIONS][obs_id])
        if obs_length == -1:  # media length not available
            interval = cfg.TIME_EVENTS

        if interval == cfg.TIME_FULL_OBS:
            min_time = 0.0
            max_time = float(obs_length)

        if interval == cfg.TIME_OBS_INTERVAL:
            obs_interval = self.pj[cfg.OBSERVATIONS][obs_id].get(cfg.OBSERVATION_TIME_INTERVAL, [0, 0])
            offset = float(self.pj[cfg.OBSERVATIONS][obs_id][cfg.TIME_OFFSET])
            min_time = float(obs_interval[0]) + offset
            # Use max media duration for max time if no interval is defined (=0)
            max_time = float(obs_interval[1]) + offset if obs_interval[1] != 0 else float(obs_length)

        if interval == cfg.TIME_EVENTS:
            try:
                min_time = float(self.pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][0][0])  # first event
            except Exception:
                min_time = 0.0
            try:
                max_time = float(self.pj[cfg.OBSERVATIONS][obs_id][cfg.EVENTS][-1][0])  # last event
            except Exception:
                max_time = float(obs_length)

        if interval == cfg.TIME_ARBITRARY_INTERVAL:
            min_time = float(start_time)
            max_time = float(end_time)

        # adjust start if start < init
        cursor.execute(
            "UPDATE aggregated_events SET start = ? WHERE  start < ? AND stop BETWEEN ? AND ?",
            (
                min_time,
                min_time,
                min_time,
                max_time,
            ),
        )
        # adjust stop if stop > end
        cursor.execute(
            "UPDATE aggregated_events SET stop = ? WHERE stop > ? AND start BETWEEN ? AND ?",
            (
                max_time,
                max_time,
                min_time,
                max_time,
            ),
        )
        # adjust start and stop if start < init and stop > end
        cursor.execute(
            "UPDATE aggregated_events SET start = ?, stop = ? WHERE start < ? AND stop > ?",
            (
                min_time,
                max_time,
                min_time,
                max_time,
            ),
        )

        # delete events outside time interval
        cursor.execute(
            "DELETE FROM aggregated_events WHERE (start < ? AND stop < ?) OR (start > ? AND stop > ?)",
            (min_time, min_time, max_time, max_time),
        )

        ylabels = [" ".join(x) for x in distinct_behav_modif]
        for ax_idx, subject in enumerate(selected_subjects):
            if parameters["exclude behaviors"]:
                cursor.execute(
                    "SELECT DISTINCT behavior, modifiers FROM aggregated_events WHERE subject = ?",
                    (subject,),
                )
                distinct_behav_modif = [[rows["behavior"], rows["modifiers"]] for rows in cursor.fetchall()]

                # add selected behaviors that are not observed
                if not parameters["exclude behaviors"]:
                    for behav in selected_behaviors:
                        if [x for x in distinct_behav_modif if x[0] == behav] == []:
                            distinct_behav_modif.append([behav, "-"])

                distinct_behav_modif = sorted(distinct_behav_modif)
                max_len = len(distinct_behav_modif)
                ylabels = [" ".join(x) for x in distinct_behav_modif]

            if not ax_idx:
                axs[ax_idx].set_title(f"Observation {obs_id}\n{subject}", fontsize=14)
            else:
                axs[ax_idx].set_title(subject, fontsize=14)
            bars = {}
            i = 0
            for behavior_modifiers in distinct_behav_modif:
                behavior, modifiers = behavior_modifiers
                behavior_modifiers_str = "|".join(behavior_modifiers) if modifiers else behavior
                bars[behavior_modifiers_str] = []

                # total duration
                cursor.execute(
                    ("SELECT start, stop FROM aggregated_events WHERE  subject = ? AND behavior = ? AND modifiers = ?"),
                    (
                        subject,
                        behavior,
                        modifiers,
                    ),
                )
                for row in cursor.fetchall():
                    bars[behavior_modifiers_str].append((row["start"], row["stop"]))

                    if self.timeFormat == cfg.HHMMSS:
                        start_date = matplotlib.dates.date2num(epoch_date + dt.timedelta(seconds=row["start"]))
                        end_date = matplotlib.dates.date2num(
                            epoch_date + dt.timedelta(seconds=row["stop"] + cfg.POINT_EVENT_PLOT_DURATION * (row["stop"] == row["start"]))
                        )
                    if self.timeFormat == cfg.S:
                        start_date = row["start"]
                        end_date = row["stop"]

                    # color
                    behav_idx = [k for k in self.pj[cfg.ETHOGRAM] if self.pj[cfg.ETHOGRAM][k]["code"] == behavior][0]
                    col = None
                    if cfg.COLOR in self.pj[cfg.ETHOGRAM][behav_idx]:
                        col = util.behavior_user_color(self.pj[cfg.ETHOGRAM], behavior)
                    if col is not None:
                        bar_color = col
                    else:
                        try:
                            bar_color = util.behavior_color(plot_colors, all_behaviors.index(behavior))
                        except Exception:
                            bar_color = "darkgray"
                    bar_color = cfg.POINT_EVENT_PLOT_COLOR if row["stop"] == row["start"] else bar_color

                    # sage colors removed from matplotlib colors list
                    if bar_color in ("sage", "darksage", "lightsage"):
                        bar_color = {
                            "darksage": "#598556",
                            "lightsage": "#bcecac",
                            "sage": "#87ae73",
                        }[bar_color]

                    try:
                        axs[ax_idx].barh(
                            (i * par1) + par1,
                            end_date - start_date,
                            left=start_date,
                            height=bar_height,
                            align="center",
                            edgecolor=bar_color,
                            color=bar_color,
                            alpha=1,
                        )
                    except Exception:
                        axs[ax_idx].barh(
                            (i * par1) + par1,
                            end_date - start_date,
                            left=start_date,
                            height=bar_height,
                            align="center",
                            edgecolor="darkgray",
                            color="darkgray",
                            alpha=1,
                        )

                i += 1

            axs[ax_idx].set_ylim(bottom=0, top=(max_len * par1) + par1)
            pos = np.arange(par1, max_len * par1 + par1 + 1, par1)
            axs[ax_idx].set_yticks(pos[: len(ylabels)])

            axs[ax_idx].set_yticklabels(ylabels, fontdict={"fontsize": 10})

            axs[ax_idx].set_ylabel(
                "Behaviors" + " (modifiers)" * include_modifiers,
                fontdict={"fontsize": 10},
            )

            if self.timeFormat == cfg.HHMMSS:
                axs[ax_idx].set_xlim(
                    left=matplotlib.dates.date2num(epoch_date + dt.timedelta(seconds=min_time)),
                    right=matplotlib.dates.date2num(epoch_date + dt.timedelta(seconds=max_time)),
                )

            axs[ax_idx].grid(color="g", linestyle=":")
            if self.timeFormat == cfg.HHMMSS:
                axs[ax_idx].xaxis_date()
                axs[ax_idx].xaxis.set_major_formatter(DateFormatter("%H:%M:%S"))
                axs[ax_idx].set_xlabel("Time (HH:MM:SS)", fontdict={"fontsize": 12})
            if self.timeFormat == cfg.S:
                axs[ax_idx].set_xlabel("Time (s)", fontdict={"fontsize": 12})

            axs[ax_idx].invert_yaxis()

        if self.timeFormat == cfg.HHMMSS:
            fig.autofmt_xdate()

        plt.tight_layout()

        if len(selected_observations) > 1:
            plt.savefig(f"{pl.Path(plot_directory) / util.safeFileName(obs_id)}.{file_format}")
        else:
            plt.show()



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'project.ui'
##
## Created by: Qt User Interface Compiler version 6.8.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QDateTimeEdit,
    QDialog, QFrame, QHBoxLayout, QHeaderView,
    QLabel, QLineEdit, QPlainTextEdit, QPushButton,
    QRadioButton, QSizePolicy, QSpacerItem, QTabWidget,
    QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget)

class Ui_dlgProject(object):
    def setupUi(self, dlgProject):
        if not dlgProject.objectName():
            dlgProject.setObjectName(u"dlgProject")
        dlgProject.resize(1202, 697)
        self.verticalLayout_7 = QVBoxLayout(dlgProject)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_6 = QVBoxLayout()
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.tabProject = QTabWidget(dlgProject)
        self.tabProject.setObjectName(u"tabProject")
        self.tabInformation = QWidget()
        self.tabInformation.setObjectName(u"tabInformation")
        self.verticalLayout = QVBoxLayout(self.tabInformation)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.horizontalLayout_15 = QHBoxLayout()
        self.horizontalLayout_15.setObjectName(u"horizontalLayout_15")
        self.label = QLabel(self.tabInformation)
        self.label.setObjectName(u"label")

        self.horizontalLayout_15.addWidget(self.label)

        self.leProjectName = QLineEdit(self.tabInformation)
        self.leProjectName.setObjectName(u"leProjectName")

        self.horizontalLayout_15.addWidget(self.leProjectName)


        self.verticalLayout.addLayout(self.horizontalLayout_15)

        self.lbProjectFilePath = QLabel(self.tabInformation)
        self.lbProjectFilePath.setObjectName(u"lbProjectFilePath")

        self.verticalLayout.addWidget(self.lbProjectFilePath)

        self.horizontalLayout_18 = QHBoxLayout()
        self.horizontalLayout_18.setObjectName(u"horizontalLayout_18")
        self.label_7 = QLabel(self.tabInformation)
        self.label_7.setObjectName(u"label_7")

        self.horizontalLayout_18.addWidget(self.label_7)

        self.dteDate = QDateTimeEdit(self.tabInformation)
        self.dteDate.setObjectName(u"dteDate")
        self.dteDate.setCalendarPopup(True)

        self.horizontalLayout_18.addWidget(self.dteDate)

        self.horizontalSpacer_10 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_18.addItem(self.horizontalSpacer_10)


        self.verticalLayout.addLayout(self.horizontalLayout_18)

        self.label_6 = QLabel(self.tabInformation)
        self.label_6.setObjectName(u"label_6")
        self.label_6.setAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)

        self.verticalLayout.addWidget(self.label_6)

        self.teDescription = QPlainTextEdit(self.tabInformation)
        self.teDescription.setObjectName(u"teDescription")

        self.verticalLayout.addWidget(self.teDescription)

        self.horizontalLayout_19 = QHBoxLayout()
        self.horizontalLayout_19.setObjectName(u"horizontalLayout_19")
        self.lbTimeFormat = QLabel(self.tabInformation)
        self.lbTimeFormat.setObjectName(u"lbTimeFormat")

        self.horizontalLayout_19.addWidget(self.lbTimeFormat)

        self.rbSeconds = QRadioButton(self.tabInformation)
        self.rbSeconds.setObjectName(u"rbSeconds")
        self.rbSeconds.setChecked(True)

        self.horizontalLayout_19.addWidget(self.rbSeconds)

        self.rbHMS = QRadioButton(self.tabInformation)
        self.rbHMS.setObjectName(u"rbHMS")

        self.horizontalLayout_19.addWidget(self.rbHMS)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_19.addItem(self.horizontalSpacer_2)


        self.verticalLayout.addLayout(self.horizontalLayout_19)

        self.lb_project_format_version = QLabel(self.tabInformation)
        self.lb_project_format_version.setObjectName(u"lb_project_format_version")

        self.verticalLayout.addWidget(self.lb_project_format_version)

        self.tabProject.addTab(self.tabInformation, "")
        self.tabEthogram = QWidget()
        self.tabEthogram.setObjectName(u"tabEthogram")
        self.verticalLayout_10 = QVBoxLayout(self.tabEthogram)
        self.verticalLayout_10.setObjectName(u"verticalLayout_10")
        self.verticalLayout_5 = QVBoxLayout()
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.horizontalLayout_11 = QHBoxLayout()
        self.horizontalLayout_11.setObjectName(u"horizontalLayout_11")
        self.twBehaviors = QTableWidget(self.tabEthogram)
        if (self.twBehaviors.columnCount() < 9):
            self.twBehaviors.setColumnCount(9)
        __qtablewidgetitem = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(0, __qtablewidgetitem)
        __qtablewidgetitem1 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(1, __qtablewidgetitem1)
        __qtablewidgetitem2 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(2, __qtablewidgetitem2)
        __qtablewidgetitem3 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(3, __qtablewidgetitem3)
        __qtablewidgetitem4 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(4, __qtablewidgetitem4)
        __qtablewidgetitem5 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(5, __qtablewidgetitem5)
        __qtablewidgetitem6 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(6, __qtablewidgetitem6)
        __qtablewidgetitem7 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(7, __qtablewidgetitem7)
        __qtablewidgetitem8 = QTableWidgetItem()
        self.twBehaviors.setHorizontalHeaderItem(8, __qtablewidgetitem8)
        self.twBehaviors.setObjectName(u"twBehaviors")
        self.twBehaviors.setAutoFillBackground(False)
        self.twBehaviors.setFrameShadow(QFrame.Sunken)
        self.twBehaviors.setMidLineWidth(0)
        self.twBehaviors.setAlternatingRowColors(True)
        self.twBehaviors.setSelectionMode(QAbstractItemView.SingleSelection)
        self.twBehaviors.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.twBehaviors.setSortingEnabled(False)
        self.twBehaviors.horizontalHeader().setProperty(u"showSortIndicator", False)
        self.twBehaviors.verticalHeader().setProperty(u"showSortIndicator", False)

        self.horizontalLayout_11.addWidget(self.twBehaviors)

        self.verticalLayout_11 = QVBoxLayout()
        self.verticalLayout_11.setObjectName(u"verticalLayout_11")
        self.pb_behavior = QPushButton(self.tabEthogram)
        self.pb_behavior.setObjectName(u"pb_behavior")

        self.verticalLayout_11.addWidget(self.pb_behavior)

        self.pb_import = QPushButton(self.tabEthogram)
        self.pb_import.setObjectName(u"pb_import")

        self.verticalLayout_11.addWidget(self.pb_import)

        self.pbBehaviorsCategories = QPushButton(self.tabEthogram)
        self.pbBehaviorsCategories.setObjectName(u"pbBehaviorsCategories")

        self.verticalLayout_11.addWidget(self.pbBehaviorsCategories)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_11.addItem(self.verticalSpacer_3)

        self.pb_exclusion_matrix = QPushButton(self.tabEthogram)
        self.pb_exclusion_matrix.setObjectName(u"pb_exclusion_matrix")

        self.verticalLayout_11.addWidget(self.pb_exclusion_matrix)

        self.pbExportEthogram = QPushButton(self.tabEthogram)
        self.pbExportEthogram.setObjectName(u"pbExportEthogram")

        self.verticalLayout_11.addWidget(self.pbExportEthogram)


        self.horizontalLayout_11.addLayout(self.verticalLayout_11)


        self.verticalLayout_5.addLayout(self.horizontalLayout_11)

        self.lbObservationsState = QLabel(self.tabEthogram)
        self.lbObservationsState.setObjectName(u"lbObservationsState")

        self.verticalLayout_5.addWidget(self.lbObservationsState)


        self.verticalLayout_10.addLayout(self.verticalLayout_5)

        self.tabProject.addTab(self.tabEthogram, "")
        self.tabSubjects = QWidget()
        self.tabSubjects.setObjectName(u"tabSubjects")
        self.verticalLayout_16 = QVBoxLayout(self.tabSubjects)
        self.verticalLayout_16.setObjectName(u"verticalLayout_16")
        self.verticalLayout_14 = QVBoxLayout()
        self.verticalLayout_14.setObjectName(u"verticalLayout_14")
        self.horizontalLayout_12 = QHBoxLayout()
        self.horizontalLayout_12.setObjectName(u"horizontalLayout_12")
        self.twSubjects = QTableWidget(self.tabSubjects)
        if (self.twSubjects.columnCount() < 3):
            self.twSubjects.setColumnCount(3)
        __qtablewidgetitem9 = QTableWidgetItem()
        self.twSubjects.setHorizontalHeaderItem(0, __qtablewidgetitem9)
        __qtablewidgetitem10 = QTableWidgetItem()
        self.twSubjects.setHorizontalHeaderItem(1, __qtablewidgetitem10)
        __qtablewidgetitem11 = QTableWidgetItem()
        self.twSubjects.setHorizontalHeaderItem(2, __qtablewidgetitem11)
        self.twSubjects.setObjectName(u"twSubjects")
        self.twSubjects.setAutoFillBackground(False)
        self.twSubjects.setFrameShadow(QFrame.Sunken)
        self.twSubjects.setMidLineWidth(0)
        self.twSubjects.setSelectionMode(QAbstractItemView.SingleSelection)
        self.twSubjects.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.twSubjects.setSortingEnabled(False)

        self.horizontalLayout_12.addWidget(self.twSubjects)

        self.verticalLayout_15 = QVBoxLayout()
        self.verticalLayout_15.setObjectName(u"verticalLayout_15")
        self.pb_subjects = QPushButton(self.tabSubjects)
        self.pb_subjects.setObjectName(u"pb_subjects")

        self.verticalLayout_15.addWidget(self.pb_subjects)

        self.pbImportSubjectsFromProject = QPushButton(self.tabSubjects)
        self.pbImportSubjectsFromProject.setObjectName(u"pbImportSubjectsFromProject")

        self.verticalLayout_15.addWidget(self.pbImportSubjectsFromProject)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_15.addItem(self.verticalSpacer_2)

        self.pb_export_subjects = QPushButton(self.tabSubjects)
        self.pb_export_subjects.setObjectName(u"pb_export_subjects")

        self.verticalLayout_15.addWidget(self.pb_export_subjects)


        self.horizontalLayout_12.addLayout(self.verticalLayout_15)


        self.verticalLayout_14.addLayout(self.horizontalLayout_12)

        self.lbSubjectsState = QLabel(self.tabSubjects)
        self.lbSubjectsState.setObjectName(u"lbSubjectsState")

        self.verticalLayout_14.addWidget(self.lbSubjectsState)


        self.verticalLayout_16.addLayout(self.verticalLayout_14)

        self.tabProject.addTab(self.tabSubjects, "")
        self.tabIndependentVariables = QWidget()
        self.tabIndependentVariables.setObjectName(u"tabIndependentVariables")
        self.horizontalLayout_14 = QHBoxLayout(self.tabIndependentVariables)
        self.horizontalLayout_14.setObjectName(u"horizontalLayout_14")
        self.horizontalLayout_13 = QHBoxLayout()
        self.horizontalLayout_13.setObjectName(u"horizontalLayout_13")
        self.verticalLayout_2 = QVBoxLayout()
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.twVariables = QTableWidget(self.tabIndependentVariables)
        if (self.twVariables.columnCount() < 5):
            self.twVariables.setColumnCount(5)
        __qtablewidgetitem12 = QTableWidgetItem()
        self.twVariables.setHorizontalHeaderItem(0, __qtablewidgetitem12)
        __qtablewidgetitem13 = QTableWidgetItem()
        self.twVariables.setHorizontalHeaderItem(1, __qtablewidgetitem13)
        __qtablewidgetitem14 = QTableWidgetItem()
        self.twVariables.setHorizontalHeaderItem(2, __qtablewidgetitem14)
        __qtablewidgetitem15 = QTableWidgetItem()
        self.twVariables.setHorizontalHeaderItem(3, __qtablewidgetitem15)
        __qtablewidgetitem16 = QTableWidgetItem()
        self.twVariables.setHorizontalHeaderItem(4, __qtablewidgetitem16)
        self.twVariables.setObjectName(u"twVariables")
        self.twVariables.setAutoFillBackground(False)
        self.twVariables.setFrameShadow(QFrame.Sunken)
        self.twVariables.setMidLineWidth(0)
        self.twVariables.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.twVariables.setDragDropOverwriteMode(False)
        self.twVariables.setAlternatingRowColors(True)
        self.twVariables.setSelectionMode(QAbstractItemView.SingleSelection)
        self.twVariables.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.twVariables.setSortingEnabled(False)
        self.twVariables.horizontalHeader().setProperty(u"showSortIndicator", False)

        self.verticalLayout_2.addWidget(self.twVariables)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.label_2 = QLabel(self.tabIndependentVariables)
        self.label_2.setObjectName(u"label_2")
        self.label_2.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_3.addWidget(self.label_2)

        self.leLabel = QLineEdit(self.tabIndependentVariables)
        self.leLabel.setObjectName(u"leLabel")

        self.horizontalLayout_3.addWidget(self.leLabel)

        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_3.addItem(self.horizontalSpacer_3)


        self.verticalLayout_2.addLayout(self.horizontalLayout_3)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.label_3 = QLabel(self.tabIndependentVariables)
        self.label_3.setObjectName(u"label_3")
        self.label_3.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_5.addWidget(self.label_3)

        self.leDescription = QLineEdit(self.tabIndependentVariables)
        self.leDescription.setObjectName(u"leDescription")

        self.horizontalLayout_5.addWidget(self.leDescription)

        self.horizontalSpacer_4 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_4)


        self.verticalLayout_2.addLayout(self.horizontalLayout_5)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.label_8 = QLabel(self.tabIndependentVariables)
        self.label_8.setObjectName(u"label_8")
        self.label_8.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_6.addWidget(self.label_8)

        self.cbType = QComboBox(self.tabIndependentVariables)
        self.cbType.setObjectName(u"cbType")
        self.cbType.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_6.addWidget(self.cbType)

        self.horizontalSpacer_5 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_6.addItem(self.horizontalSpacer_5)


        self.verticalLayout_2.addLayout(self.horizontalLayout_6)

        self.horizontalLayout_7 = QHBoxLayout()
        self.horizontalLayout_7.setObjectName(u"horizontalLayout_7")
        self.label_4 = QLabel(self.tabIndependentVariables)
        self.label_4.setObjectName(u"label_4")
        self.label_4.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_7.addWidget(self.label_4)

        self.lePredefined = QLineEdit(self.tabIndependentVariables)
        self.lePredefined.setObjectName(u"lePredefined")

        self.horizontalLayout_7.addWidget(self.lePredefined)

        self.horizontalSpacer_6 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_7.addItem(self.horizontalSpacer_6)


        self.verticalLayout_2.addLayout(self.horizontalLayout_7)

        self.horizontalLayout_8 = QHBoxLayout()
        self.horizontalLayout_8.setObjectName(u"horizontalLayout_8")
        self.label_9 = QLabel(self.tabIndependentVariables)
        self.label_9.setObjectName(u"label_9")
        self.label_9.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_8.addWidget(self.label_9)

        self.dte_default_date = QDateTimeEdit(self.tabIndependentVariables)
        self.dte_default_date.setObjectName(u"dte_default_date")

        self.horizontalLayout_8.addWidget(self.dte_default_date)

        self.horizontalSpacer_7 = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_8.addItem(self.horizontalSpacer_7)


        self.verticalLayout_2.addLayout(self.horizontalLayout_8)

        self.horizontalLayout_9 = QHBoxLayout()
        self.horizontalLayout_9.setObjectName(u"horizontalLayout_9")
        self.label_5 = QLabel(self.tabIndependentVariables)
        self.label_5.setObjectName(u"label_5")

        self.horizontalLayout_9.addWidget(self.label_5)

        self.leSetValues = QLineEdit(self.tabIndependentVariables)
        self.leSetValues.setObjectName(u"leSetValues")

        self.horizontalLayout_9.addWidget(self.leSetValues)

        self.horizontalSpacer_8 = QSpacerItem(40, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_9.addItem(self.horizontalSpacer_8)


        self.verticalLayout_2.addLayout(self.horizontalLayout_9)


        self.horizontalLayout_13.addLayout(self.verticalLayout_2)

        self.verticalLayout_12 = QVBoxLayout()
        self.verticalLayout_12.setObjectName(u"verticalLayout_12")
        self.pbAddVariable = QPushButton(self.tabIndependentVariables)
        self.pbAddVariable.setObjectName(u"pbAddVariable")

        self.verticalLayout_12.addWidget(self.pbAddVariable)

        self.pbRemoveVariable = QPushButton(self.tabIndependentVariables)
        self.pbRemoveVariable.setObjectName(u"pbRemoveVariable")

        self.verticalLayout_12.addWidget(self.pbRemoveVariable)

        self.pbImportVarFromProject = QPushButton(self.tabIndependentVariables)
        self.pbImportVarFromProject.setObjectName(u"pbImportVarFromProject")

        self.verticalLayout_12.addWidget(self.pbImportVarFromProject)

        self.verticalSpacer_4 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_12.addItem(self.verticalSpacer_4)


        self.horizontalLayout_13.addLayout(self.verticalLayout_12)


        self.horizontalLayout_14.addLayout(self.horizontalLayout_13)

        self.tabProject.addTab(self.tabIndependentVariables, "")
        self.tabBehavCodingMap = QWidget()
        self.tabBehavCodingMap.setObjectName(u"tabBehavCodingMap")
        self.verticalLayout_8 = QVBoxLayout(self.tabBehavCodingMap)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.twBehavCodingMap = QTableWidget(self.tabBehavCodingMap)
        if (self.twBehavCodingMap.columnCount() < 2):
            self.twBehavCodingMap.setColumnCount(2)
        __qtablewidgetitem17 = QTableWidgetItem()
        self.twBehavCodingMap.setHorizontalHeaderItem(0, __qtablewidgetitem17)
        __qtablewidgetitem18 = QTableWidgetItem()
        self.twBehavCodingMap.setHorizontalHeaderItem(1, __qtablewidgetitem18)
        self.twBehavCodingMap.setObjectName(u"twBehavCodingMap")
        self.twBehavCodingMap.setAutoFillBackground(False)
        self.twBehavCodingMap.setFrameShadow(QFrame.Sunken)
        self.twBehavCodingMap.setMidLineWidth(0)
        self.twBehavCodingMap.setSelectionMode(QAbstractItemView.SingleSelection)
        self.twBehavCodingMap.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.twBehavCodingMap.setSortingEnabled(False)

        self.horizontalLayout.addWidget(self.twBehavCodingMap)

        self.verticalLayout_4 = QVBoxLayout()
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.pbAddBehaviorsCodingMap = QPushButton(self.tabBehavCodingMap)
        self.pbAddBehaviorsCodingMap.setObjectName(u"pbAddBehaviorsCodingMap")

        self.verticalLayout_4.addWidget(self.pbAddBehaviorsCodingMap)

        self.pbRemoveBehaviorsCodingMap = QPushButton(self.tabBehavCodingMap)
        self.pbRemoveBehaviorsCodingMap.setObjectName(u"pbRemoveBehaviorsCodingMap")

        self.verticalLayout_4.addWidget(self.pbRemoveBehaviorsCodingMap)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_4.addItem(self.verticalSpacer)


        self.horizontalLayout.addLayout(self.verticalLayout_4)


        self.verticalLayout_8.addLayout(self.horizontalLayout)

        self.tabProject.addTab(self.tabBehavCodingMap, "")
        self.tab_time_converters = QWidget()
        self.tab_time_converters.setObjectName(u"tab_time_converters")
        self.verticalLayout_18 = QVBoxLayout(self.tab_time_converters)
        self.verticalLayout_18.setObjectName(u"verticalLayout_18")
        self.label_11 = QLabel(self.tab_time_converters)
        self.label_11.setObjectName(u"label_11")

        self.verticalLayout_18.addWidget(self.label_11)

        self.horizontalLayout_16 = QHBoxLayout()
        self.horizontalLayout_16.setObjectName(u"horizontalLayout_16")
        self.tw_converters = QTableWidget(self.tab_time_converters)
        if (self.tw_converters.columnCount() < 3):
            self.tw_converters.setColumnCount(3)
        __qtablewidgetitem19 = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(0, __qtablewidgetitem19)
        __qtablewidgetitem20 = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(1, __qtablewidgetitem20)
        __qtablewidgetitem21 = QTableWidgetItem()
        self.tw_converters.setHorizontalHeaderItem(2, __qtablewidgetitem21)
        self.tw_converters.setObjectName(u"tw_converters")
        self.tw_converters.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.tw_converters.setSelectionMode(QAbstractItemView.SingleSelection)
        self.tw_converters.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.tw_converters.setSortingEnabled(False)

        self.horizontalLayout_16.addWidget(self.tw_converters)

        self.verticalLayout_17 = QVBoxLayout()
        self.verticalLayout_17.setObjectName(u"verticalLayout_17")
        self.pb_add_converter = QPushButton(self.tab_time_converters)
        self.pb_add_converter.setObjectName(u"pb_add_converter")

        self.verticalLayout_17.addWidget(self.pb_add_converter)

        self.pb_modify_converter = QPushButton(self.tab_time_converters)
        self.pb_modify_converter.setObjectName(u"pb_modify_converter")

        self.verticalLayout_17.addWidget(self.pb_modify_converter)

        self.pb_delete_converter = QPushButton(self.tab_time_converters)
        self.pb_delete_converter.setObjectName(u"pb_delete_converter")

        self.verticalLayout_17.addWidget(self.pb_delete_converter)

        self.pb_load_from_file = QPushButton(self.tab_time_converters)
        self.pb_load_from_file.setObjectName(u"pb_load_from_file")

        self.verticalLayout_17.addWidget(self.pb_load_from_file)

        self.pb_load_from_repo = QPushButton(self.tab_time_converters)
        self.pb_load_from_repo.setObjectName(u"pb_load_from_repo")

        self.verticalLayout_17.addWidget(self.pb_load_from_repo)

        self.verticalSpacer_7 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_17.addItem(self.verticalSpacer_7)


        self.horizontalLayout_16.addLayout(self.verticalLayout_17)


        self.verticalLayout_18.addLayout(self.horizontalLayout_16)

        self.horizontalLayout_17 = QHBoxLayout()
        self.horizontalLayout_17.setObjectName(u"horizontalLayout_17")
        self.label_13 = QLabel(self.tab_time_converters)
        self.label_13.setObjectName(u"label_13")
        self.label_13.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_17.addWidget(self.label_13)

        self.le_converter_name = QLineEdit(self.tab_time_converters)
        self.le_converter_name.setObjectName(u"le_converter_name")

        self.horizontalLayout_17.addWidget(self.le_converter_name)

        self.horizontalSpacer_11 = QSpacerItem(10, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_17.addItem(self.horizontalSpacer_11)


        self.verticalLayout_18.addLayout(self.horizontalLayout_17)

        self.horizontalLayout_10 = QHBoxLayout()
        self.horizontalLayout_10.setObjectName(u"horizontalLayout_10")
        self.label_10 = QLabel(self.tab_time_converters)
        self.label_10.setObjectName(u"label_10")
        self.label_10.setMinimumSize(QSize(120, 0))

        self.horizontalLayout_10.addWidget(self.label_10)

        self.le_converter_description = QLineEdit(self.tab_time_converters)
        self.le_converter_description.setObjectName(u"le_converter_description")

        self.horizontalLayout_10.addWidget(self.le_converter_description)

        self.horizontalSpacer_9 = QSpacerItem(10, 20, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_10.addItem(self.horizontalSpacer_9)


        self.verticalLayout_18.addLayout(self.horizontalLayout_10)

        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.verticalLayout_9 = QVBoxLayout()
        self.verticalLayout_9.setObjectName(u"verticalLayout_9")
        self.label_12 = QLabel(self.tab_time_converters)
        self.label_12.setObjectName(u"label_12")

        self.verticalLayout_9.addWidget(self.label_12)

        self.pb_code_help = QPushButton(self.tab_time_converters)
        self.pb_code_help.setObjectName(u"pb_code_help")

        self.verticalLayout_9.addWidget(self.pb_code_help)

        self.verticalSpacer_5 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_9.addItem(self.verticalSpacer_5)


        self.horizontalLayout_2.addLayout(self.verticalLayout_9)

        self.pteCode = QPlainTextEdit(self.tab_time_converters)
        self.pteCode.setObjectName(u"pteCode")
        font = QFont()
        font.setFamilies([u"Monospace"])
        self.pteCode.setFont(font)

        self.horizontalLayout_2.addWidget(self.pteCode)

        self.verticalLayout_13 = QVBoxLayout()
        self.verticalLayout_13.setObjectName(u"verticalLayout_13")
        self.pb_save_converter = QPushButton(self.tab_time_converters)
        self.pb_save_converter.setObjectName(u"pb_save_converter")

        self.verticalLayout_13.addWidget(self.pb_save_converter)

        self.pb_cancel_converter = QPushButton(self.tab_time_converters)
        self.pb_cancel_converter.setObjectName(u"pb_cancel_converter")

        self.verticalLayout_13.addWidget(self.pb_cancel_converter)

        self.verticalSpacer_6 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_13.addItem(self.verticalSpacer_6)


        self.horizontalLayout_2.addLayout(self.verticalLayout_13)


        self.verticalLayout_18.addLayout(self.horizontalLayout_2)

        self.tabProject.addTab(self.tab_time_converters, "")

        self.verticalLayout_6.addWidget(self.tabProject)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout_4.addItem(self.horizontalSpacer)

        self.pbCancel = QPushButton(dlgProject)
        self.pbCancel.setObjectName(u"pbCancel")

        self.horizontalLayout_4.addWidget(self.pbCancel)

        self.pbOK = QPushButton(dlgProject)
        self.pbOK.setObjectName(u"pbOK")

        self.horizontalLayout_4.addWidget(self.pbOK)


        self.verticalLayout_6.addLayout(self.horizontalLayout_4)


        self.verticalLayout_7.addLayout(self.verticalLayout_6)


        self.retranslateUi(dlgProject)

        self.tabProject.setCurrentIndex(3)


        QMetaObject.connectSlotsByName(dlgProject)
    # setupUi

    def retranslateUi(self, dlgProject):
        dlgProject.setWindowTitle(QCoreApplication.translate("dlgProject", u"Project", None))
        self.label.setText(QCoreApplication.translate("dlgProject", u"Project name", None))
        self.lbProjectFilePath.setText(QCoreApplication.translate("dlgProject", u"Project file path:", None))
        self.label_7.setText(QCoreApplication.translate("dlgProject", u"Project date and time", None))
        self.dteDate.setDisplayFormat(QCoreApplication.translate("dlgProject", u"yyyy-MM-dd hh:mm", None))
        self.label_6.setText(QCoreApplication.translate("dlgProject", u"Project description", None))
        self.lbTimeFormat.setText(QCoreApplication.translate("dlgProject", u"Project time format", None))
        self.rbSeconds.setText(QCoreApplication.translate("dlgProject", u"seconds", None))
        self.rbHMS.setText(QCoreApplication.translate("dlgProject", u"hh:mm:ss.mss", None))
        self.lb_project_format_version.setText(QCoreApplication.translate("dlgProject", u"Project format version:", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tabInformation), QCoreApplication.translate("dlgProject", u"Information", None))
        ___qtablewidgetitem = self.twBehaviors.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("dlgProject", u"Behavior type", None));
        ___qtablewidgetitem1 = self.twBehaviors.horizontalHeaderItem(1)
        ___qtablewidgetitem1.setText(QCoreApplication.translate("dlgProject", u"Key", None));
        ___qtablewidgetitem2 = self.twBehaviors.horizontalHeaderItem(2)
        ___qtablewidgetitem2.setText(QCoreApplication.translate("dlgProject", u"Code", None));
        ___qtablewidgetitem3 = self.twBehaviors.horizontalHeaderItem(3)
        ___qtablewidgetitem3.setText(QCoreApplication.translate("dlgProject", u"Description", None));
        ___qtablewidgetitem4 = self.twBehaviors.horizontalHeaderItem(4)
        ___qtablewidgetitem4.setText(QCoreApplication.translate("dlgProject", u"Color", None));
        ___qtablewidgetitem5 = self.twBehaviors.horizontalHeaderItem(5)
        ___qtablewidgetitem5.setText(QCoreApplication.translate("dlgProject", u"Category", None));
        ___qtablewidgetitem6 = self.twBehaviors.horizontalHeaderItem(6)
        ___qtablewidgetitem6.setText(QCoreApplication.translate("dlgProject", u"Modifiers", None));
        ___qtablewidgetitem7 = self.twBehaviors.horizontalHeaderItem(7)
        ___qtablewidgetitem7.setText(QCoreApplication.translate("dlgProject", u"Exclusion", None));
        ___qtablewidgetitem8 = self.twBehaviors.horizontalHeaderItem(8)
        ___qtablewidgetitem8.setText(QCoreApplication.translate("dlgProject", u"Modifiers coding map", None));
        self.pb_behavior.setText(QCoreApplication.translate("dlgProject", u"Behavior", None))
        self.pb_import.setText(QCoreApplication.translate("dlgProject", u"Import ethogram", None))
        self.pbBehaviorsCategories.setText(QCoreApplication.translate("dlgProject", u"Behavioral categories", None))
        self.pb_exclusion_matrix.setText(QCoreApplication.translate("dlgProject", u"Exclusion matrix", None))
        self.pbExportEthogram.setText(QCoreApplication.translate("dlgProject", u"Export ethogram", None))
        self.lbObservationsState.setText(QCoreApplication.translate("dlgProject", u"TextLabel", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tabEthogram), QCoreApplication.translate("dlgProject", u"Ethogram", None))
        ___qtablewidgetitem9 = self.twSubjects.horizontalHeaderItem(0)
        ___qtablewidgetitem9.setText(QCoreApplication.translate("dlgProject", u"Key", None));
        ___qtablewidgetitem10 = self.twSubjects.horizontalHeaderItem(1)
        ___qtablewidgetitem10.setText(QCoreApplication.translate("dlgProject", u"Subject name", None));
        ___qtablewidgetitem11 = self.twSubjects.horizontalHeaderItem(2)
        ___qtablewidgetitem11.setText(QCoreApplication.translate("dlgProject", u"Description", None));
        self.pb_subjects.setText(QCoreApplication.translate("dlgProject", u"Subjects", None))
        self.pbImportSubjectsFromProject.setText(QCoreApplication.translate("dlgProject", u"Import subjects", None))
        self.pb_export_subjects.setText(QCoreApplication.translate("dlgProject", u"Export subjects", None))
        self.lbSubjectsState.setText(QCoreApplication.translate("dlgProject", u"TextLabel", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tabSubjects), QCoreApplication.translate("dlgProject", u"Subjects", None))
        ___qtablewidgetitem12 = self.twVariables.horizontalHeaderItem(0)
        ___qtablewidgetitem12.setText(QCoreApplication.translate("dlgProject", u"Label", None));
        ___qtablewidgetitem13 = self.twVariables.horizontalHeaderItem(1)
        ___qtablewidgetitem13.setText(QCoreApplication.translate("dlgProject", u"Description", None));
        ___qtablewidgetitem14 = self.twVariables.horizontalHeaderItem(2)
        ___qtablewidgetitem14.setText(QCoreApplication.translate("dlgProject", u"Type", None));
        ___qtablewidgetitem15 = self.twVariables.horizontalHeaderItem(3)
        ___qtablewidgetitem15.setText(QCoreApplication.translate("dlgProject", u"Predefined value", None));
        ___qtablewidgetitem16 = self.twVariables.horizontalHeaderItem(4)
        ___qtablewidgetitem16.setText(QCoreApplication.translate("dlgProject", u"Set of values", None));
        self.label_2.setText(QCoreApplication.translate("dlgProject", u"Label", None))
        self.label_3.setText(QCoreApplication.translate("dlgProject", u"Description", None))
        self.label_8.setText(QCoreApplication.translate("dlgProject", u"Type", None))
        self.label_4.setText(QCoreApplication.translate("dlgProject", u"Predefined value", None))
        self.label_9.setText(QCoreApplication.translate("dlgProject", u"Predefined timestamp", None))
        self.dte_default_date.setDisplayFormat(QCoreApplication.translate("dlgProject", u"yyyy-MM-dd hh:mm:ss.zzz", None))
        self.label_5.setText(QCoreApplication.translate("dlgProject", u"Set of values (separated by comma)", None))
        self.pbAddVariable.setText(QCoreApplication.translate("dlgProject", u"Add variable", None))
        self.pbRemoveVariable.setText(QCoreApplication.translate("dlgProject", u"Remove variable", None))
        self.pbImportVarFromProject.setText(QCoreApplication.translate("dlgProject", u"Import variables\n"
"from a BORIS project", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tabIndependentVariables), QCoreApplication.translate("dlgProject", u"Independent variables", None))
        ___qtablewidgetitem17 = self.twBehavCodingMap.horizontalHeaderItem(0)
        ___qtablewidgetitem17.setText(QCoreApplication.translate("dlgProject", u"Name", None));
        ___qtablewidgetitem18 = self.twBehavCodingMap.horizontalHeaderItem(1)
        ___qtablewidgetitem18.setText(QCoreApplication.translate("dlgProject", u"Behavior codes", None));
        self.pbAddBehaviorsCodingMap.setText(QCoreApplication.translate("dlgProject", u"Add a behaviors coding map", None))
        self.pbRemoveBehaviorsCodingMap.setText(QCoreApplication.translate("dlgProject", u"Remove behaviors coding map", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tabBehavCodingMap), QCoreApplication.translate("dlgProject", u"Behaviors coding map", None))
        self.label_11.setText(QCoreApplication.translate("dlgProject", u"Time converters for external data", None))
        ___qtablewidgetitem19 = self.tw_converters.horizontalHeaderItem(0)
        ___qtablewidgetitem19.setText(QCoreApplication.translate("dlgProject", u"Name", None));
        ___qtablewidgetitem20 = self.tw_converters.horizontalHeaderItem(1)
        ___qtablewidgetitem20.setText(QCoreApplication.translate("dlgProject", u"Description", None));
        ___qtablewidgetitem21 = self.tw_converters.horizontalHeaderItem(2)
        ___qtablewidgetitem21.setText(QCoreApplication.translate("dlgProject", u"Code", None));
        self.pb_add_converter.setText(QCoreApplication.translate("dlgProject", u"Add new converter", None))
        self.pb_modify_converter.setText(QCoreApplication.translate("dlgProject", u"Modify converter", None))
        self.pb_delete_converter.setText(QCoreApplication.translate("dlgProject", u"Delete converter", None))
        self.pb_load_from_file.setText(QCoreApplication.translate("dlgProject", u"Load converters from file", None))
        self.pb_load_from_repo.setText(QCoreApplication.translate("dlgProject", u"Load converters from BORIS repository", None))
        self.label_13.setText(QCoreApplication.translate("dlgProject", u"Name", None))
        self.label_10.setText(QCoreApplication.translate("dlgProject", u"Description", None))
        self.label_12.setText(QCoreApplication.translate("dlgProject", u"Python code", None))
        self.pb_code_help.setText(QCoreApplication.translate("dlgProject", u"Help", None))
        self.pb_save_converter.setText(QCoreApplication.translate("dlgProject", u"Save converter", None))
        self.pb_cancel_converter.setText(QCoreApplication.translate("dlgProject", u"Cancel", None))
        self.tabProject.setTabText(self.tabProject.indexOf(self.tab_time_converters), QCoreApplication.translate("dlgProject", u"Converters", None))
        self.pbCancel.setText(QCoreApplication.translate("dlgProject", u"Cancel", None))
        self.pbOK.setText(QCoreApplication.translate("dlgProject", u"OK", None))
    # retranslateUi

